from typing import NamedTuple
from .erd_oven_cook_mode import ErdOvenCookMode

class AvailableCookMode(NamedTuple):
    """Parsing helper for Available Cook Modes"""
    byte: int
    mask: int
    cook_mode: ErdOvenCookMode



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import ErdPrecisionCookingProbeBatteryStatus

class ErdPrecisionCookingProbeBatteryStatusConverter(ErdReadOnlyConverter[ErdPrecisionCookingProbeBatteryStatus]):
    def erd_decode(self, value: str) -> ErdPrecisionCookingProbeBatteryStatus:
        if value:
            try:
                return ErdPrecisionCookingProbeBatteryStatus(erd_decode_int(value[:2]))  
            except ValueError:
                return ErdPrecisionCookingProbeBatteryStatus.NA
        else:
            return ErdPrecisionCookingProbeBatteryStatus.NA
        



from ..abstract import ErdReadWriteConverter, ErdReadOnlyConverter

def erd_decode_bytes(value: any) -> bytes:
    """Decode a raw bytes ERD value sent as a hex encoded string."""
    return bytes.fromhex(value)
def erd_encode_bytes(value: bytes) -> str:
    """Encode a raw bytes ERD value."""
    return value.hex('big')

class ErdBytesConverter(ErdReadWriteConverter[bytes]):
    def erd_decode(self, value: str) -> bytes:
        """Decode a raw bytes ERD value sent as a hex encoded string."""
        return erd_decode_bytes(value)
    def erd_encode(self, value: bytes) -> str:
        """Encode a raw bytes ERD value."""
        return erd_encode_bytes(value)

class ErdReadOnlyBytesConverter(ErdReadOnlyConverter[bytes]):
    def erd_decode(self, value: str) -> bytes:
        """Decode a raw bytes ERD value sent as a hex encoded string."""
        return erd_decode_bytes(value)



from typing import NamedTuple, Optional

from .const import *
from .oven_cook_mode import OvenCookMode 

class OvenCookSetting(NamedTuple):
    """Cleaner representation of ErdOvenCookMode"""
    cook_mode: OvenCookMode
    temperature: int
    raw_bytes: Optional[bytes] = None

    def stringify(self, **kwargs) -> Optional[str]:
        """Format OvenCookSetting values nicely."""
        cook_state = self.cook_mode.oven_state
        temp_units = kwargs.get("temp_units", "°")

        modifiers = []
        if self.cook_mode.timed:
            modifiers.append(STATE_OVEN_TIMED)
        if self.cook_mode.delayed:
            modifiers.append(STATE_OVEN_DELAY)
        if self.cook_mode.probe:
            modifiers.append(STATE_OVEN_PROBE)
        if self.cook_mode.sabbath:
            modifiers.append(STATE_OVEN_SABBATH)

        temp_str = f" ({self.temperature}{temp_units})" if self.temperature > 0 else ""
        modifier_str = f" ({', '.join(modifiers)})" if modifiers else ""
        display_state = cook_state.stringify(**kwargs)

        if display_state in [STATE_OVEN_UNKNOWN, STATE_OVEN_OFF]:
            return display_state
            
        return f"{display_state}{temp_str}{modifier_str}"
    def __str__(self) -> str:
        return self.stringify() or ""



import enum

@enum.unique
class ErdOperatingMode(enum.Enum):
    NA = "FF"
    PAUSE = "04"
    LOW_POWER = "00"
    POWER_UP = "01"
    STANDBY = "02"
    DELAY_START = "03"
    CYCLE_ACTIVE = "05"
    EOC = "06"
    DOWNLOAD_MODE = "07"
    SENSOR_CHECK_MODE = "08"
    LOAD_ACTIVATION_MODE = "09"
    MC_ONLY_MODE = "11"
    WARNING_MODE = "12"
    CONTROL_LOCKED = "13"
    CSM_TRIPPED = "14"




from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.fridge import FridgeModelInfo

class FridgeModelInfoConverter(ErdReadOnlyConverter[FridgeModelInfo]):
    def erd_decode(self, value: str) -> FridgeModelInfo:
        doors = 2
        has_fridge = True
        has_freezer = True
        has_convertable_drawer = False

        type_code = erd_decode_int(value)
        if(type_code == 3):
            has_freezer = False
            doors = 1
        if(type_code == 4):
            has_fridge = False
            doors = 1
        if(type_code in [6,7]):
            doors = 4
            has_convertable_drawer = True
        
        return FridgeModelInfo(
            has_fridge=has_fridge,
            has_freezer=has_freezer,
            doors=doors,
            has_convertable_drawer=has_convertable_drawer,
            raw_value=type_code
        )




import enum

@enum.unique
class ErdApplianceType(enum.Enum):
    UNKNOWN = "FF"
    WATER_HEATER = "00"
    DRYER = "01"
    WASHER = "02"
    FRIDGE = "03"
    MICROWAVE = "04"
    ADVANTIUM = "05"
    DISH_WASHER = "06"
    OVEN = "07"
    ELECTRIC_RANGE = "08"
    GAS_RANGE = "09"
    AIR_CONDITIONER = "0a"
    ELECTRIC_COOKTOP = "0b"
    COOKTOP = "11"
    PIZZA_OVEN = "0c"
    GAS_COOKTOP = "0d"
    SPLIT_AIR_CONDITIONER = "0e"
    HOOD = "0f"
    POE_WATER_FILTER = "10"
    WATER_SOFTENER = "15"
    PORTABLE_AIR_CONDITIONER = "16"
    COMBINATION_WASHER_DRYER = "17"
    ZONELINE = "14"
    DELEVERY_BOX = "12"
    CAFE_COFFEE_MAKER = "1A"




import enum

@enum.unique
class ErdOvenConfiguration(enum.Enum):
    """Representation of oven configurations."""
    HAS_KNOB = 1
    HAS_WARMING_DRAWER = 2
    HAS_LIGHT_BAR = 4
    HAS_LOWER_OVEN = 8
    HAS_LOWER_OVEN_KITCHEN_TIMER = 16



from .erd_bool_converter import erd_decode_bool, erd_encode_bool, ErdBoolConverter, ErdReadOnlyBoolConverter
from .erd_bytes_converter import erd_decode_bytes, erd_encode_bytes, ErdBytesConverter, ErdReadOnlyBytesConverter
from .erd_int_converter import erd_decode_int, erd_encode_int, ErdIntConverter, ErdReadOnlyIntConverter
from .erd_signed_byte_converter import erd_decode_signed_byte, erd_encode_signed_byte, ErdSignedByteConverter, ErdReadOnlySignedByteConverter
from .erd_string_converter import erd_decode_string, erd_encode_string, ErdStringConverter, ErdReadOnlyStringConverter
from .erd_time_span_converter import erd_decode_timespan, erd_encode_timespan, ErdTimeSpanConverter, ErdReadOnlyTimeSpanConverter



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import ErdPrecisionCookingStartSousVideTimerActiveStatus

class ErdPrecisionCookingStartSousVideTimerActiveStatusConverter(ErdReadOnlyConverter[ErdPrecisionCookingStartSousVideTimerActiveStatus]):
    def erd_decode(self, value: str) -> ErdPrecisionCookingStartSousVideTimerActiveStatus:
        if value:
            try:
                return ErdPrecisionCookingStartSousVideTimerActiveStatus(erd_decode_int(value[:2]))  
            except ValueError:
                return ErdPrecisionCookingStartSousVideTimerActiveStatus.NA
        else:
            return ErdPrecisionCookingStartSousVideTimerActiveStatus.NA



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import ErdPrecisionCookingProbeTargetTimeReached

class ErdPrecisionCookingProbeTargetTimeReachedConverter(ErdReadOnlyConverter[ErdPrecisionCookingProbeTargetTimeReached]):
    def erd_decode(self, value: str) -> ErdPrecisionCookingProbeTargetTimeReached:
        if value:
            try:
                return ErdPrecisionCookingProbeTargetTimeReached(erd_decode_int(value[:2]))  
            except ValueError:
                return ErdPrecisionCookingProbeTargetTimeReached.NA
        else:
            return ErdPrecisionCookingProbeTargetTimeReached.NA



from typing import NamedTuple, Optional
from .erd_cooktop_status import ErdCooktopStatus
from .burner import Burner

class CooktopStatus(NamedTuple):
    status: ErdCooktopStatus
    burners: dict
    raw_value: Optional[str]

    @classmethod
    def DEFAULT(cls):
        return cls(ErdCooktopStatus.DASH, {}, None)

    @property
    def left_front(self) -> Optional[Burner]:
        try:
            return self.burners["leftFront"]
        except(KeyError):
            return None

    @property
    def left_rear(self) -> Optional[Burner]:
        try:
            return self.burners["leftRear"]
        except(KeyError):
            return None
    
    @property
    def center_front(self) -> Optional[Burner]:
        try:
            return self.burners["centerFront"]
        except(KeyError):
            return None

    @property
    def center_rear(self) -> Optional[Burner]:
        try:
            return self.burners["centerRear"]
        except(KeyError):
            return None

    @property
    def right_front(self) -> Optional[Burner]:
        try:
            return self.burners["rightFront"]
        except(KeyError):
            return None
    
    @property
    def right_rear(self) -> Optional[Burner]:
        try:
            return self.burners["rightRear"]
        except(KeyError):
            return None

    def boolify(self):
        return self.status == ErdCooktopStatus.BURNERS_ON



#WIP

import enum

#0x5771
#0x5772
#0x5773
#0x5774
#0x5775
#0x5776
#0x5777

class ErdCloseLoopCookingDevicesStatus:
    class CookingDeviceType(enum.Enum):
        DEFAULT = "Default"
        HESTANT_CUE_PAN = "HestanCuePan"
        HESTANT_CUE_POT = "HestanCuePot"
        PARAGON_SENSOR = "Paragon"
        POPUP_SENSOR = "PopupSensor"

    _CODE_MAPPING = {
        "0x5771": "0x5778",
        "0x5772": "0x5779",
        "0x5773": "0x577a",
        "0x5774": "0x577b",
        "0x5775": "0x577c",
        "0x5776": "0x577d",
        "0x5777": "0x577e"
    }
    _y0 = 100
    _z0 = 450

    def __init__(self, value: str, code: str):
        self.code = ErdCloseLoopCookingDevicesStatus._CODE_MAPPING[code]
        self.t0 = -1
        self.u0 = -1

        if not value:
            self.status = "NoData"
            self.device_type = self.CookingDeviceType.DEFAULT
            self.r0 = 0
        try:
            self.device_type = self._initialize_device_type(value)
            self.r0 = int(value[2,4], 16) #TODO: based on the original logic, this needs error handling
            self.status = value[4,8]
            if len(value) >= 16:
                self.t0 = int(value[8:12], 16)
                self.u0 = int(value[12:16], 16)
            elif self.device_type == self.CookingDeviceType.POPUP_SENSOR:
                self.t0 = ErdCloseLoopCookingDevicesStatus.y0
                self.u0 = ErdCloseLoopCookingDevicesStatus.z0
                pass            
        except:
            self.device_type = self.CookingDeviceType.DEFAULT
            self.r0 = 0
                    
    def _initialize_device_type(self, value: str):
        try:
            return {
                "00": self.device_type.POPUP_SENSOR,
                "01": self.device_type.PARAGON_SENSOR,
                "02": self.device_type.HESTANT_CUE_PAN,
                "03": self.device_type.HESTANT_CUE_POT
            }[value[:2]]
        except:
            return self.device_type.DEFAULT




from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.fridge import FridgeDoorStatus, ErdDoorStatus

class FridgeDoorStatusConverter(ErdReadOnlyConverter[FridgeDoorStatus]):
    def erd_decode(self, value: str) -> FridgeDoorStatus:
        def get_door_status(val: str) -> ErdDoorStatus:
            try:
                return ErdDoorStatus(val)
            except ValueError:
                return ErdDoorStatus.NA

        fridge_right = get_door_status(value[:2])
        fridge_left = get_door_status(value[2:4])
        freezer = get_door_status(value[4:6])
        drawer = get_door_status(value[6:8])
        if (fridge_right != ErdDoorStatus.OPEN) and (fridge_left != ErdDoorStatus.OPEN):
            if freezer == ErdDoorStatus.OPEN:
                status = "Freezer Open"
            else:
                status = "Closed"
        elif freezer == ErdDoorStatus.OPEN:
            status = "All Open"
        else:
            status = "Fridge Open"
        return FridgeDoorStatus(
            fridge_right=fridge_right,
            fridge_left=fridge_left,
            freezer=freezer,
            drawer=drawer,
            status=status,
        )



import enum

@enum.unique
class ErdOvenCookMode(enum.Enum):
    """See ErdCookMode.smali"""
    BAKED_GOODS = 60
    BAKETIMEDSHUTOFF_DELAYSTART = 7
    BAKETIMED_TWOTEMP = 5
    BAKETIMED_TWOTEMP_DELAYSTART = 9
    BAKETIMED_WARM = 4
    BAKETIMED_WARM_DELAYSTART = 8
    BAKE_DELAYSTART = 3
    BAKE_NOOPTION = 1
    BAKE_PROBE = 2
    BAKE_PROBE_DELAYSTART = 6
    BAKE_SABBATH = 10
    BROIL_HIGH = 12
    BROIL_LOW = 11
    CONVBAKETIMEDSHUTOFF_DELAYSTART = 24
    CONVBAKETIMED_TWOTEMP = 22
    CONVBAKETIMED_TWOTEMP_DELAYSTART = 26
    CONVBAKETIMED_WARM = 21
    CONVBAKETIMED_WARM_DELAYSTART = 25
    CONVBAKE_DELAYSTART = 20
    CONVBAKE_NOOPTION = 18
    CONVBAKE_PROBE = 19
    CONVBAKE_PROBE_DELAYSTART = 23
    CONVBROILCRISP_NOOPTION = 47
    CONVBROILCRISP_PROBE = 48
    CONVBROIL_HIGH_NOOPTION = 46
    CONVBROIL_LOW_NOOPTION = 45
    CONVMULTIBAKETIMEDSHUTOFF_DELAYSTART = 33
    CONVMULTIBAKETIMED_TWOTEMP = 31
    CONVMULTIBAKETIMED_TWOTEMP_DELAYSTART = 35
    CONVMULTIBAKETIMED_WARM = 30
    CONVMULTIBAKETIMED_WARM_DELAYSTART = 34
    CONVMULTIBAKE_DELAYSTART = 29
    CONVMULTIBAKE_NOOPTION = 27
    CONVMULTIBAKE_PROBE = 28
    CONVMULTIBAKE_PROBE_DELAYSTART = 32
    CONVROASTTIMEDSHUTOFF_DELAYSTART = 42
    CONVROASTTIMED_TWOTEMP = 40
    CONVROASTTIMED_TWOTEMP_DELAYSTART = 44
    CONVROASTTIMED_WARM = 39
    CONVROASTTIMED_WARM_DELAYSTART = 43
    CONVROAST_DELAYSTART = 38
    CONVROAST_NOOPTION = 36
    CONVROAST_PROBE = 37
    CONVROAST_PROBE_DELAYSTART = 41
    CUSTOMSELFCLEAN = 49
    CUSTOMSELFCLEAN_DELAYSTART = 50
    DUALBROIL_HIGH_NOOPTION = 54
    DUALBROIL_LOW_NOOPTION = 53
    FROZEN_PIZZA = 58
    FROZEN_PIZZA_MULTI = 59
    FROZEN_SNACKS = 56
    FROZEN_SNACKS_MULTI = 567
    NOMODE = 0
    PROOF_DELAYSTART = 14
    PROOF_NOOPTION = 13
    STEAMCLEAN = 51
    STEAMCLEAN_DELAYSTART = 52
    WARM_DELAYSTART = 17
    WARM_NOOPTION = 15
    WARM_PROBE = 16



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import CooktopStatus, ErdCooktopStatus, Burner

class CooktopStatusConverter(ErdReadOnlyConverter[CooktopStatus]):
    def erd_decode(self, value: str) -> CooktopStatus:
        if not value:
            return ErdCooktopStatus.DEFAULT()
        
        try:
            # break the string into two character segments and parse as ints
            vals = [erd_decode_int(value[i:i + 2]) for i in range(0, len(value), 2)]
            status = ErdCooktopStatus(vals[0])
            burners = {}

            burners["leftFront"] = Burner(vals[1], vals[2])
            burners["leftRear"] = Burner(vals[3], vals[4])
            burners["centerFront"] = Burner(vals[5], vals[6])
            burners["centerRear"] = Burner(vals[7], vals[8])
            burners["rightFront"] = Burner(vals[9], vals[10])
            burners["rightRear"] = Burner(vals[11], vals[12])

            return CooktopStatus(status, burners, value)
        except:
            return CooktopStatus.DEFAULT()



import asyncio
import os
import logging

import slixmpp
import socket
import ssl
from aiohttp import ClientSession
from lxml import etree
from ..const import (
    EVENT_ADD_APPLIANCE,
    EVENT_APPLIANCE_STATE_CHANGE,
)
from ..async_login_flow import async_do_full_xmpp_flow
from ..erd import ErdCode, ErdCodeType
from ..ge_appliance import GeAppliance
from .base_client import GeBaseClient
from typing import Any, Dict, List, Optional, Set, Tuple, Union, Callable

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)

# If this isn't done, it'll throw a not implemented exception
if os.name == 'nt':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

def _first_or_none(lst: list) -> Any:
    try:
        return lst[0]
    except IndexError:
        return None


set_timer = False


def _format_request(
        msg_id: Union[int, str], uri: str, method: str, key: Optional[str] = None, value: Optional[str] = None) -> str:
    """Format a XMPP pseudo-HTTP request."""
    if method.lower() == 'post':
        post_body = f"<json>{json.dumps({key:value})}</json>"
    else:
        post_body = ""
    message = (
        f"<request><id>{msg_id}</id>"
        f"<method>{method}</method>"
        f"<uri>{uri}</uri>"
        f"{post_body}"
        "</request>"  # [sic.]
    )
    return message


class GeXmppClient(GeBaseClient, slixmpp.ClientXMPP):
    client_priority = 1

    def add_event_handler(self, event: str, callback: Callable, disposable: bool = False):
        super(GeBaseClient, self).add_event_handler(event, callback, disposable)

    def __init__(self, xmpp_credentials: Dict, event_loop: Optional[asyncio.AbstractEventLoop] = None,
                 known_jids: Optional[List[str]] = None, **kwargs):
        GeBaseClient.__init__(self, event_loop)
        slixmpp.ClientXMPP.__init__(self, xmpp_credentials['jid'], xmpp_credentials['password'], **kwargs)
        self.credentials = xmpp_credentials
        self.register_plugin('xep_0030')  # Service Discovery
        self.register_plugin('xep_0199')  # XMPP Ping

        self.add_event_handler('message', self.on_message)
        self.add_event_handler('session_start', self.on_start)
        self.add_event_handler('presence_available', self.on_presence_available)
        self.add_event_handler('presence_unavailable', self.on_presence_unavailable)
        self.add_event_handler(EVENT_APPLIANCE_STATE_CHANGE, self.maybe_trigger_appliance_init_event)
        self.ssl_context.verify_mode = ssl.CERT_NONE

        if known_jids is None:
            known_jids = []
        self._known_jids = [self.complete_jid(j) for j in known_jids]

        self._initial_update_ids = set()  # type: Set[str]

    def connect(self, address: Optional[Tuple[str, int]] = None, use_ssl: bool = False,
                force_starttls: bool = True, disable_starttls: bool = False):
        """Connect to the XMPP server."""
        if address is None:
            address = (self._credentials['address'], self._credentials['port'])
        super().connect(
            address=address,
            use_ssl=use_ssl,
            force_starttls=force_starttls,
            disable_starttls=disable_starttls
        )

    async def add_appliance(self, jid: str):
        """Add an appliance to the registry and request an update."""
        mac_addr = jid.split('_')[0]
        if jid in self.appliances:
            # TODO: Make this a package-specific exception
            raise RuntimeError('Trying to add duplicate appliance')
        new_appliance = GeAppliance(mac_addr, self)

        _LOGGER.info(f'Adding appliance {jid}')
        self.appliances[jid] = new_appliance
        self.event(EVENT_ADD_APPLIANCE, new_appliance)
        await new_appliance.async_request_update()

    async def maybe_add_add_appliance(self, jid: str):
        """Add an appliance, suppressing the error if it already exists."""
        try:
            await self.add_appliance(jid)
        except RuntimeError:
            pass

    async def on_presence_available(self, evt: slixmpp.ElementBase):
        """Perform actions when notified of an available JID."""
        await asyncio.sleep(2)  # Wait 2 seconds to give it time to register
        jid = slixmpp.JID(evt['from']).bare

        if jid == self.boundjid.bare:
            return
        try:
            self.appliances[jid].set_available()
            _LOGGER.debug(f'Appliance {jid} marked available')
        except KeyError:
            await self.add_appliance(jid)
            self.appliances[jid].set_available()

    async def on_presence_unavailable(self, evt):
        """When appliance is no longer available, mark it as such."""
        jid = slixmpp.JID(evt['from']).bare

        if jid == self.boundjid.bare:
            return
        try:
            self.appliances[jid].set_unavailable()
            _LOGGER.debug(f'Appliance {jid} marked unavailable')
        except KeyError:
            pass

    async def on_message(self, event):
        """Handle incoming messages."""
        global set_timer
        msg = str(event)
        msg_from = slixmpp.JID(event['from']).bare
        try:
            message_data = self.extract_message_json(msg)
        except ValueError:
            _LOGGER.info(f"From: {msg_from}: Not a GE message")
            return
        try:
            appliance = self.appliances[msg_from]
            state_changes = appliance.update_erd_values(message_data)
            if state_changes:
                self.event(EVENT_APPLIANCE_STATE_CHANGE, [appliance, state_changes])
        except KeyError:
            _LOGGER.warning('Received json message from unregistered appliance')

    async def on_start(self, event):
        _LOGGER.debug('Starting session: ' + str(event))
        self.send_presence('available')
        await asyncio.sleep(5)
        for jid in self._known_jids:
            await self.maybe_add_add_appliance(jid)

    def complete_jid(self, jid) -> str:
        """Make a complete jid from a username"""
        if "@" in jid:
            return jid
        return f"{jid}@{self.boundjid.host}"

    def get_appliance_jid(self, appliance: GeAppliance) -> str:
        return self.complete_jid(f"{appliance.mac_addr}_{self.user_id}")

    @staticmethod
    def extract_message_json(message: str) -> Dict:
        """The appliances send messages that don't play nice with slixmpp, so let's do this."""
        etr = etree.XML(message)
        json_elem = _first_or_none(etr.xpath('//json'))
        if json_elem is None:
            raise ValueError('Not a GE appliance message')

        data = _first_or_none(json_elem.xpath('text()'))
        data = json.loads(data)
        return data

    async def process_in_running_loop(self, timeout: Optional[int] = None):
        """Process all the available XMPP events (receiving or sending data on the
        socket(s), calling various registered callbacks, calling expired
        timers, handling signal events, etc).  If timeout is None, this
        coroutine will run forever or until disconnected. If timeout is a number,
        this function will return after the given time in seconds."""

        if timeout is None:
            await asyncio.ensure_future(self.disconnected, loop=self.loop)
        else:
            await asyncio.wait([self.disconnected], loop=self.loop, timeout=timeout)

    async def _connect_routine(self):
        """
        Override the _connect_routine method from xmlstream.py.
        This is to address the bug corrected in open PR https://github.com/poezio/slixmpp/pull/19.
        """
        self.event_when_connected = "connected"

        if self.connect_loop_wait > 0:
            self.event('reconnect_delay', self.connect_loop_wait)
            await asyncio.sleep(self.connect_loop_wait, loop=self.loop)

        record = await self.pick_dns_answer(self.default_domain)
        if record is not None:
            host, address, dns_port = record
            port = self.address[1] if self.address[1] else dns_port
            self.address = (address, port)
            self._service_name = host
        else:
            # No DNS records left, stop iterating
            # and try (host, port) as a last resort
            self.dns_answers = None

        if self.use_ssl:
            ssl_context = self.get_ssl_context()
        else:
            ssl_context = None

        if self._current_connection_attempt is None:
            return
        try:
            await self.loop.create_connection(
                lambda: self, self.address[0], self.address[1], ssl=ssl_context,
                server_hostname=self.default_domain if self.use_ssl else None
            )
            self.connect_loop_wait = 0
        except socket.gaierror:
            self.event('connection_failed', 'No DNS record available for %s' % self.default_domain)
        except OSError as e:
            _LOGGER.debug('Connection failed: %s', e)
            self.event("connection_failed", e)
            if self._current_connection_attempt is None:
                return
            self.connect_loop_wait = self.connect_loop_wait * 2 + 1
            self._current_connection_attempt = asyncio.ensure_future(
                self._connect_routine(), loop=self.loop,
            )

    def send_raw_message(self, mto: slixmpp.JID, mbody: str, mtype: str = 'chat', msg_id: Optional[str] = None):
        """TODO: Use actual xml for this instead of hacking it.  Then again, this is what GE does in the app."""
        if msg_id is None:
            msg_id = self.new_id()
        raw_message = (
            f'<message type="{mtype}" from="{self.boundjid.bare}" to="{mto}" id="{msg_id}">'
            f'<body>{mbody}</body>'
            f'</message>'
        )
        self.send_raw(raw_message)

    def send_request(
            self, appliance: GeAppliance, method: str, uri: str, key: Optional[str] = None,
            value: Optional[str] = None, message_id: Optional[str] = None):
        """
        Send a pseudo-http request to the appliance
        :param appliance: GeAppliance, the appliance to send the request to
        :param method: str, Usually "GET" or "POST"
        :param uri: str, the "endpoint" for the request, usually of the form "/UUID/erd/{erd_code}"
        :param key: The json key to set in a POST request.  Usually a four-character hex string with leading "0x".
        :param value: The value to set, usually encoded as a hex string without a leading "0x"
        :param message_id:
        """
        if method.lower() != 'post' and (key is not None or value is not None):
            raise RuntimeError('Values can only be set in a POST request')

        if message_id is None:
            message_id = self.new_id()
        message_body = _format_request(message_id, uri, method, key, value)
        jid = slixmpp.JID(self.get_appliance_jid(appliance))
        self.send_raw_message(jid, message_body)

    async def async_request_update(self, appliance: GeAppliance):
        """Request a full update from the appliance."""
        self.send_request(appliance, 'GET', '/UUID/cache')

    async def async_set_erd_value(self, appliance: GeAppliance, erd_code: ErdCodeType, erd_value: Any):
        """
        Send a new erd value to the appliance
        :param appliance: GeAppliance, the appliance to update
        :param erd_code: The ERD code to update
        :param erd_value: The new value to set
        """
        if isinstance(erd_code, ErdCode):
            raw_erd_code = erd_code.value
        else:
            raw_erd_code = erd_code

        uri = f'/UUID/erd/{raw_erd_code}'
        self.send_request(appliance, 'POST', uri, raw_erd_code, erd_value)

    async def async_get_credentials(self, session: ClientSession, username: str, password: str) -> bool:
        xmpp_credentials = await async_do_full_xmpp_flow(session, username, password)
        self.credentials = xmpp_credentials
        self.set_jid(xmpp_credentials['jid'])
        self.password = xmpp_credentials['password']
        return True

    async def async_event(self, event: str, data: Any = None):
        """Make awaiting work here"""
        super(GeBaseClient, self).event(event, data)



import enum
from typing import Optional

@enum.unique
class ErdFilterStatus(enum.Enum):
    GOOD = "00"
    REPLACE = "01"
    EXPIRED = "02"
    UNFILTERED = "03"
    LEAK_DETECTED = "04"
    NA = "FF"

    def stringify(self, **kwargs) -> Optional[str]:
        if(self.value == ErdFilterStatus.NA):
            return "N/A"
        return self.name.title()



from .erd_cycle_state import ErdCycleState, ErdCycleStateRaw

CYCLE_STATE_RAW_MAP = {
    ErdCycleStateRaw.STATE_01: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.STATE_02: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.STATE_03: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.STATE_04: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.STATE_05: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.STATE_06: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.STATE_07: ErdCycleState.SENSING,
    ErdCycleStateRaw.STATE_08: ErdCycleState.MAIN_WASH,
    ErdCycleStateRaw.STATE_09: ErdCycleState.MAIN_WASH,
    ErdCycleStateRaw.STATE_10: ErdCycleState.DRYING,
    ErdCycleStateRaw.STATE_11: ErdCycleState.SANITIZING,
    ErdCycleStateRaw.STATE_12: ErdCycleState.RINSING,
    ErdCycleStateRaw.STATE_13: ErdCycleState.RINSING,
    ErdCycleStateRaw.STATE_14: ErdCycleState.RINSING,
    ErdCycleStateRaw.STATE_15: ErdCycleState.RINSING,
    ErdCycleStateRaw.STATE_16: ErdCycleState.PAUSE,
    ErdCycleStateRaw.STATE_17: ErdCycleState.NA,
    ErdCycleStateRaw.STATE_18: ErdCycleState.NA
}



from .erd_rinse_agent import ErdRinseAgent, ErdRinseAgentRaw

RINSE_AGENT_RAW_MAP = {
    ErdRinseAgentRaw.RINSE_AGENT_GOOD: ErdRinseAgent.RINSE_AGENT_GOOD,
    ErdRinseAgentRaw.RINSE_AGENT_LOW1: ErdRinseAgent.RINSE_AGENT_LOW,
    ErdRinseAgentRaw.RINSE_AGENT_LOW2: ErdRinseAgent.RINSE_AGENT_LOW
}



from gekitchensdk.erd.erd_code_class import ErdCodeClass
import logging
from typing import Any

from ..exception import GeUnsupportedOperationError

from .erd_configuration import ErdConfigurationEntry, _configuration
from .erd_codes import ErdCode, ErdCodeType
from .converters import *

_LOGGER = logging.getLogger(__name__)

class ErdEncoder:
    def __init__(self) -> None:
        super().__init__()
        self._registry = dict((k.erd_code, k) for k in _configuration)

    def translate_code(self, erd_code: ErdCodeType) -> ErdCodeType:
        """
        Try to resolve an ERD codes from string to ErdCode if possible.  If an ErdCode
        object is passed in, it will be returned.
        :param erd_code: ErdCode or str
        :return: Either an ErdCode object matching the `erd_code` string, or, if resolution fails,
        the `erd_code` string itself.
        """
        if isinstance(erd_code, ErdCode):
            return erd_code

        try:
            return ErdCode[erd_code]
        except KeyError:
            pass

        try:
            return ErdCode(erd_code.lower())
        except ValueError:
            return erd_code

    def decode_value(self, erd_code: ErdCodeType, erd_value: str) -> Any:
        """
        Decode and ERD Code raw value into something useful.  If the erd_code is a string that
        cannot be resolved to a known ERD Code, the value will be treated as raw byte string.
        Unregistered ERD Codes will be translated as ints.

        :param erd_code: ErdCode or str, the ERD Code the value of which we want to decode
        :param erd_value: The raw ERD code value, usually a hex string without leading "0x"
        :return: The decoded value.
        """
        if erd_value == '':
            return None

        erd_code = self.translate_code(erd_code)

        if isinstance(erd_code, str):
            return erd_decode_bytes(erd_value)

        try:
            return self._registry[erd_code].erd_decode(erd_value)
        except KeyError:
            return erd_decode_int(erd_value)

    def get_code_class(self, erd_code: ErdCodeType) -> ErdCodeClass:
        """
        Gets the code class for a given ErdCode.  Returns GENERAL if not
        available.
        """
        erd_code = self.translate_code(erd_code)
        if isinstance(erd_code, str):
            return ErdCodeClass.GENERAL
        
        try:
            return self._registry[erd_code].code_class
        except KeyError:
            return ErdCodeClass.GENERAL
    
    def encode_value(self, erd_code: ErdCodeType, value: Any) -> str:
        """
        Encode an ERD Code value as a hex string.
        Only ERD Codes registered with self.erd_encoders will processed.  Otherwise an error will be returned.

        :param erd_code: ErdCode or str, the ERD Code the value of which we want to decode
        :param value: The value to re-encode
        :return: The encoded value as a hex string
        """
        if value is None:
            return ''

        erd_code = self.translate_code(erd_code)

        try:
            return self._registry[erd_code].erd_encode(value)
        except KeyError:
            _LOGGER.error(f'Attempt to encode unregistered ERD code {erd_code}')
            raise

    def can_decode(self, erd_code: ErdCodeType) -> bool:
        """ 
        Indicates whether an ERD Code can be decoded. If the code
        is not registered, defaults to true
        """

        erd_code = self.translate_code(erd_code)

        try:
            return self._registry[erd_code].can_decode
        except KeyError:
            return True

    def can_encode(self, erd_code: ErdCodeType) -> bool:
        """ 
        Indicates whether an ERD Code can be encoded. If the code
        is not registered, defaults to false
        """

        erd_code = self.translate_code(erd_code)

        try:
            return self._registry[erd_code].can_encode
        except KeyError:
            return False

    def can_boolify(self, erd_code: ErdCodeType) -> bool:
        """
        Indicates whether an ERD Code can boolified.  If the code
        is not registered, defaults to false
        """

        erd_code = self.translate_code(erd_code)

        try:
            return self._registry[erd_code].can_boolify
        except KeyError:
            return False
        


from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import ErdAvailableCookMode

class ErdAvailableCookModeConverter(ErdReadOnlyConverter[ErdAvailableCookMode]):
    def erd_decode(self, value: str) -> ErdAvailableCookMode:
        if not value:
            return {ErdAvailableCookMode.OVEN_BAKE.value.cook_mode}
        mode_bytes = [int(i) for i in erd_decode_bytes(value)]
        available_modes = {
            mode.value.cook_mode
            for mode in ErdAvailableCookMode
            if mode_bytes[mode.value.byte] & mode.value.mask
        }
        return available_modes



from typing import NamedTuple
from .erd_oven_state import ErdOvenState

class OvenCookMode(NamedTuple):
    """Named tuple to represent ErdOvenCookMode for easier formatting later"""
    oven_state: ErdOvenState
    delayed: bool = False
    timed: bool = False
    probe: bool = False
    warm: bool = False
    sabbath: bool = False



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import ErdOvenState


class ErdOvenStateConverter(ErdReadOnlyConverter[ErdOvenState]):
    def erd_decode(self, value: str) -> ErdOvenState:
        """
        See erdCurrentState.smali
        """
        state_code = erd_decode_int(value)
        if 44 <= state_code <= 59:
            return ErdOvenState.OVEN_STATE_SPECIAL_X
        if 42 <= state_code <= 43:
            return ErdOvenState.OVEN_STATE_BAKED_GOODS
        if 40 <= state_code <= 41:
            return ErdOvenState.OVEN_STATE_FROZEN_PIZZA_MULTI
        if 38 <= state_code <= 39:
            return ErdOvenState.OVEN_STATE_FROZEN_SNACKS_MULTI
        if 36 <= state_code <= 37:
            return ErdOvenState.OVEN_STATE_FROZEN_PIZZA
        if 33 <= state_code <= 35:
            return ErdOvenState.OVEN_STATE_FROZEN_SNACKS
        if 1 <= state_code <= 27:
            # These 27 were nicely enumerated in v1.0.3 of the app, though the display logic is more similar to
            # to those above, grouping similar things. See ErdCurrentState.smali for more details.
            return ErdOvenState(state_code)
        return ErdOvenState.STATUS_DASH



from ..abstract import ErdReadWriteConverter
from ..primitives import *
from gekitchensdk.erd.values.common import ErdOnOff
from gekitchensdk.erd.values.fridge import IceMakerControlStatus

class IceMakerControlStatusConverter(ErdReadWriteConverter[IceMakerControlStatus]):
    def erd_decode(self, value: str) -> IceMakerControlStatus:
        def parse_status(val: str) -> ErdOnOff:
            try:
                return ErdOnOff(val)
            except ValueError:
                return ErdOnOff.NA

        status_fz = parse_status(value[:2])
        status_ff = parse_status(value[2:])

        return IceMakerControlStatus(status_fridge=status_ff, status_freezer=status_fz)
    def erd_encode(self, value: IceMakerControlStatus):
        return value.status_freezer.value + value.status_fridge.value



import enum

@enum.unique
class OperatingMode(enum.Enum):
    STATUS_DASH = "status_dash"
    STATUS_CYCLE_ACTIVE = "status_active"
    STATUS_OFF = "status_off"
    STATUS_DELAY = "status_delay"
    STATUS_PAUSED = "status_paused"
    CONTROL_LOCKED = "control_locked"
    
    def stringify(self, **kwargs):
        return self.name.replace("STATUS_","").replace("_"," ").title()



"""Async HTTP authentication flow to get XMPP loggin credentials"""
# TODO: Refactor the two login flows for DRYness

try:
    import re2 as re
except ImportError:
    import re
import logging
import aiohttp
from lxml import etree
from .const import (
    API_URL,
    LOGIN_URL,
    OAUTH2_APP_ID,
    OAUTH2_CLIENT_ID,
    OAUTH2_CLIENT_SECRET,
    OAUTH2_REDIRECT_URI,
)
from .exception import GeAuthFailedError, GeGeneralServerError

from typing import Dict
from urllib.parse import urlparse, parse_qs

_LOGGER = logging.getLogger(__name__)


async def async_get_oauth2_token(session: aiohttp.ClientSession, username: str, password: str) -> Dict:
    """Hackily get an oauth2 token until I can be bothered to do this correctly"""
    params = {
        'client_id': OAUTH2_CLIENT_ID,
        'response_type': 'code',
        'access_type': 'offline',
        'redirect_uri': OAUTH2_REDIRECT_URI,
    }

    async with session.get(f'{LOGIN_URL}/oauth2/auth', params=params) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        resp_text = await resp.text()

    email_regex = (
        r'^\s*(\w+(?:(?:-\w+)|(?:\.\w+)|(?:\+\w+))*\@'
        r'[A-Za-z0-9]+(?:(?:\.|-)[A-Za-z0-9]+)*\.[A-Za-z0-9][A-Za-z0-9]+)\s*$'
    )
    clean_username = re.sub(email_regex, r'\1', username)

    etr = etree.HTML(resp_text)
    post_data = {
        i.attrib['name']: i.attrib['value']
        for i in etr.xpath("//form[@id = 'frmsignin']//input")
        if 'value' in i.keys()
    }
    post_data['username'] = clean_username
    post_data['password'] = password

    async with session.post(f'{LOGIN_URL}/oauth2/g_authenticate', data=post_data, allow_redirects=False) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        code = parse_qs(urlparse(resp.headers['Location']).query)['code'][0]

    post_data = {
        'code': code,
        'client_id': OAUTH2_CLIENT_ID,
        'client_secret': OAUTH2_CLIENT_SECRET,
        'redirect_uri': OAUTH2_REDIRECT_URI,
        'grant_type': 'authorization_code',
    }
    auth = aiohttp.BasicAuth(OAUTH2_CLIENT_ID, OAUTH2_CLIENT_SECRET)
    async with session.post(f'{LOGIN_URL}/oauth2/token', data=post_data, auth=auth) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        oauth_token = await resp.json()
    try:
        return {'Authorization': 'Bearer ' + oauth_token['access_token']}
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a token: {oauth_token}')


async def async_get_mobile_device_token(session: aiohttp.ClientSession, auth_header: Dict) -> str:
    """Get a mobile device token"""
    mdt_data = {
        'kind': 'mdt#login',
        'app': OAUTH2_APP_ID,
        'os': 'google_android'
    }
    async with session.post(f'{API_URL}/v1/mdt', json=mdt_data, headers=auth_header) as resp:
        if resp.status != 200:
            raise GeAuthFailedError(await resp.text())
        results = await resp.json()

    try:
        return results['mdt']
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a mobile device token: {results}')


async def async_get_ge_token(session: aiohttp.ClientSession, auth_header: Dict, mobile_device_token: str) -> str:
    """Get the GE token that we can use to get XMPP credentials"""
    params = {
        'client_id': OAUTH2_CLIENT_ID,
        'client_secret': OAUTH2_CLIENT_SECRET,
        'mdt': mobile_device_token
    }
    async with session.post(f'{LOGIN_URL}/oauth2/getoken', params=params, headers=auth_header) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        results = await resp.json()

    try:
        return results['access_token']
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a GE token: {results}')


async def async_get_xmpp_credentials(session: aiohttp.ClientSession, ge_token: str) -> Dict:
    """Get XMPP credentials"""
    uri = f'{API_URL}/v1/mdt/credentials'
    headers = {'Authorization': f'Bearer {ge_token}'}
    async with session.get(uri, headers=headers) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        return await resp.json()


async def async_get_wss_credentials(session: aiohttp.ClientSession, auth_header: Dict) -> Dict:
    """Get WSS credentials"""
    uri = f'{API_URL}/v1/websocket'
    async with session.get(uri, headers=auth_header) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        return await resp.json()


async def async_do_full_xmpp_flow(session: aiohttp.ClientSession, username: str, password: str) -> Dict:
    """Perform a complete login flow, returning XMPP credentials."""

    _LOGGER.debug('Getting oauth2 token')
    auth_header = await async_get_oauth2_token(session, username, password)

    _LOGGER.debug('Getting mobile device token')
    mdt = await async_get_mobile_device_token(session, auth_header)

    _LOGGER.debug('Getting GE token')
    ge_token = await async_get_ge_token(session, auth_header, mdt)

    _LOGGER.debug('Getting XMPP credentials')
    xmpp_credentials = await async_get_xmpp_credentials(session, ge_token)

    return xmpp_credentials


async def async_do_full_wss_flow(session: aiohttp.ClientSession, username: str, password: str) -> Dict:
    """Perform a complete login flow, returning WSS credentials."""

    _LOGGER.debug('Getting oauth2 token')
    auth_header = await async_get_oauth2_token(session, username, password)

    _LOGGER.debug('Getting mobile device token')
    wss_credentials = await async_get_wss_credentials(session, auth_header)

    return wss_credentials




from ..abstract import ErdReadOnlyConverter
from ..primitives import *

from gekitchensdk.erd.values import ErdApplianceType

class ErdApplianceTypeConverter(ErdReadOnlyConverter[ErdApplianceType]):
    def erd_decode(self, value) -> ErdApplianceType:
        try:
            return ErdApplianceType(value)
        except ValueError:
            return ErdApplianceType.UNKNOWN



from .erd_available_cook_mode_converter import ErdAvailableCookModeConverter
from .erd_cooktop_config_converter import ErdCooktopConfigConverter
from .erd_oven_state_converter import ErdOvenStateConverter
from .erd_precision_cooking_app_probe_control_mode_converter import ErdPrecisionCookingAppProbeControlModeConverter
from .erd_precision_cooking_probe_battery_status_converter import ErdPrecisionCookingProbeBatteryStatusConverter
from .erd_precision_cooking_probe_target_time_reached_converter import ErdPrecisionCookingProbeTargetTimeReachedConverter
from .erd_precision_cooking_start_sous_vide_timer_active_status_converter import ErdPrecisionCookingStartSousVideTimerActiveStatusConverter

from .cooktop_status_converter import CooktopStatusConverter
from .oven_configuration_converter import OvenConfigurationConverter
from .oven_cook_mode_converter import OvenCookModeConverter
from .oven_ranges_converter import OvenRangesConverter



import asyncio
import logging
import websockets
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple

from aiohttp import ClientSession

from .base_client import GeBaseClient

from ..async_login_flow import async_do_full_wss_flow
from ..exception import GeNotAuthenticatedError
from ..const import (
    API_URL,
    EVENT_ADD_APPLIANCE,
    EVENT_APPLIANCE_STATE_CHANGE,
    EVENT_APPLIANCE_UPDATE_RECEIVED,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_GOT_APPLIANCE_LIST,
)
from ..erd import ErdCode, ErdCodeType
from ..ge_appliance import GeAppliance
from ..login_flow import do_full_wss_flow

try:
    import ujson as json
except ImportError:
    import json

ALL_ERD = "allErd"
API_HOST = API_URL[8:]  # Drop the https://
CALLBACK_TIMEOUT = 0.05
LIST_APPLIANCES = "List-appliances"
SET_ERD = "setErd"
_LOGGER = logging.getLogger(__name__)


class GeWebsocketClient(GeBaseClient):
    """
    Client for GE's Websocket pseudo-MQTT API.

    TODO:
      Handle re-auth
      Better error handling in general
      Probably a ton of other stuff?
    """
    client_priority = 2  # This should be the primary client

    def __init__(self, event_loop: Optional[asyncio.AbstractEventLoop] = None,
                 username: Optional[str] = None, password: Optional[str] = None):
        super().__init__(event_loop)
        self._endpoint = None  # type: Optional[str]
        self._socket = None  # type: Optional[websockets.client.WebSocketClientProtocol]
        self._pending_erds = {}  # type: Dict[Tuple[str, str], str]
        self._event_handlers = defaultdict(list)  # type: Dict[str, List[Callable]]
        self.username = username
        self.password = password
        self._keepalive_fut = None  # type: Optional[asyncio.Future]

        self.add_event_handler(EVENT_APPLIANCE_STATE_CHANGE, self.maybe_trigger_appliance_init_event)

    def _apply_default_login(self, username: Optional[str] = None, password: Optional[str] = None) -> Tuple[str, str]:
        if username is None:
            username = self.username
        if password is None:
            password = self.password
        if username is None or password is None:
            raise RuntimeError('username and password must be specified')
        return username, password

    async def async_get_credentials(self, session: ClientSession, username: Optional[str] = None,
                                    password: Optional[str] = None):
        username, password = self._apply_default_login(username, password)
        wss_credentials = await async_do_full_wss_flow(session, username, password)
        self.credentials = wss_credentials

    def get_credentials(self, username: Optional[str] = None, password: Optional[str] = None):
        username, password = self._apply_default_login(username, password)
        wss_credentials = do_full_wss_flow(username, password)
        self.credentials = wss_credentials

    @property
    def endpoint(self) -> str:
        try:
            return self.credentials['endpoint']
        except (TypeError, KeyError):
            raise GeNotAuthenticatedError

    @property
    def event_handlers(self) -> Dict[str, List[Callable]]:
        return self._event_handlers

    def add_event_handler(self, event: str, callback: Callable, disposable: bool = False):
        if disposable:
            raise NotImplementedError('Support for disposable callbacks not yet implemented')
        self.event_handlers[event].append(callback)

    async def async_event(self, event: str, *args, **kwargs):
        """Trigger event callbacks sequentially"""
        for cb in self.event_handlers[event]:
            asyncio.ensure_future(cb(*args, **kwargs), loop=self.loop)

    @property
    def websocket(self):
        return self._socket

    async def _process_pending_erd(self, message_id: str):
        id_parts = message_id.split("-")
        if id_parts[1] != SET_ERD:
            raise ValueError("Invalid message id")
        mac_addr = id_parts[0]
        raw_erd_code = id_parts[2]
        erd_value = self._pending_erds.get((mac_addr, raw_erd_code))
        if erd_value is not None:
            _LOGGER.debug(f"")
            try:
                await self._update_appliance_state(mac_addr, {raw_erd_code: erd_value})
            except KeyError:
                pass

    @staticmethod
    async def _failed_set_erd(message_dict: dict):
        """Failed to set an ERD value."""
        body = message_dict.get("body", {})
        message = body.get("message")
        reason = body.get("reason")
        _LOGGER.error(f"ERD update failure!\nMessage: {message}.\nReason: {reason}")

    async def process_message(self, message: str):
        """
        Process an incoming message.
        """
        message_dict = json.loads(message)  # type: Dict
        try:
            kind = message_dict['kind']
        except KeyError:
            return

        if kind.lower() == "publish#erd":
            await self.process_erd_update(message_dict)
        elif kind.lower() == "websocket#api":
            try:
                message_id = message_dict["id"]
            except KeyError:
                return
            if message_id == LIST_APPLIANCES:
                await self.process_appliance_list(message_dict)
            elif f"-{SET_ERD}-" in message_id and message_dict.get("code") == 200:
                await self._process_pending_erd(message_id)
            elif f"-{SET_ERD}-" in message_id:
                await self._failed_set_erd(message_dict)
            elif f"-{ALL_ERD}" in message_id and message_dict.get("code") == 200:
                await self.process_cache_update(message_dict)

    async def process_appliance_list(self, message_dict: Dict):
        """
        Process the appliance list.

        These messages should take the form::

            {"kind": "websocket#api",
             "id": "List-appliances",
             "request":{...},
             "success": True,
             "code": 200,
             "body": {
                "kind": "appliance#applianceList",
                "userId": "USER_ID",
                "items":[
                    {"applianceId": "MAC_ADDR_1",
                     "type": "TYPE_1",
                     "brand": "Unknown",
                     "jid":"<MAC_ADDR_1>_<USER_ID>",
                     "nickname":"NICKNAME",
                     "online":"ONLINE"
                    },
                    ...,
                ],
            }

        :param message_dict:
        """
        body = message_dict["body"]
        if body.get("kind") != "appliance#applianceList":
            raise ValueError("Not an applianceList")
        items = body["items"]
        for item in items:
            mac_addr = item["applianceId"].upper()
            if mac_addr in self.appliances:
                continue
            online = item['online'].upper() == "ONLINE"
            await self.add_appliance(mac_addr, online)
        await self.async_event(EVENT_GOT_APPLIANCE_LIST, items)

    async def process_cache_update(self, message_dict: Dict):
        """
        Process an appliance's full cache update.
        
        These messages should take the form::

            {
                "body": {
                    "applianceId": "MAC_ADDRESS",
                    "items": [
                        {"erd": "ERD_CODE_1", "time": "UPDATE_TIMESTAMP_1", "value": "VALUE_1"},
                        ...,
                        {"erd": "ERD_CODE_N", "time": "UPDATE_TIMESTAMP_N", "value": "VALUE_N"},
                    ],
                    "kind": "appliance#erdList",
                    "userId": "USER_ID"
                },
                "code": 200,
                "id": "<MAC_ADDRESS>-allErd",
                "kind": "websocket#api",
                "request": {...},
                "success": True,
            }
        """
        body = message_dict["body"]
        if body.get("kind") != "appliance#erdList":
            raise ValueError("Not an erdList")
        mac_addr = body["applianceId"].upper()
        updates = {i["erd"]: i["value"] for i in body["items"]}
        await self._update_appliance_state(mac_addr, updates)

    async def process_erd_update(self, message_dict: Dict):
        """
        Process an ERD update (pseudo-HTTP PUBLISH).

        These messages should be in the form::

            {
                "item": {
                    "applianceId": "MAC_ADDRESS",
                    "erd": "ERD_CODE",
                    "time": "UPDATE_TIMESTAMP",
                    "value": "SOME_VALUE",
                },
                "resource": "/appliance/<MAC_ADDRESS>/erd/<ERD_CODE>",
                "kind": "publish#erd",
                "userId":"USER_ID"
            }

        :param message_dict: dict, the json-decoded message.
        """
        item = message_dict["item"]
        mac_addr = item["applianceId"].upper()
        update = {item['erd']: item['value']}
        await self._update_appliance_state(mac_addr, update)

    async def _update_appliance_state(self, mac_addr: str, updates: Dict[ErdCodeType, str]):
        """Update appliance state, performing callbacks if necessary."""
        try:
            appliance = self.appliances[mac_addr]
        except KeyError:
            return
        state_changes = appliance.update_erd_values(updates)
        if state_changes:
            await self.async_event(EVENT_APPLIANCE_STATE_CHANGE, [appliance, state_changes])
        await self.async_event(EVENT_APPLIANCE_UPDATE_RECEIVED, [appliance, updates])

    async def disconnect(self):
        """Disconnect and cleanup."""
        _LOGGER.info("Disconnecting")
        await self.websocket.close()
        await self.async_event(EVENT_DISCONNECTED, self._socket)

    async def async_get_credentials_and_run(
            self, session: ClientSession, username: Optional[str] = None, password: Optional[str] = None,
            appliances: Optional[List[GeAppliance]] = None, keepalive: Optional[int] = 30):
        """Do a full login flow and run the client."""
        await self.async_get_credentials(session, username, password)
        await self.async_run_client(appliances, keepalive)

    async def async_run_client(self, appliances: Optional[List[GeAppliance]] = None, keepalive: Optional[int] = 30):
        """Run the client."""
        async with websockets.connect(self.endpoint) as socket:
            self._socket = socket
            if keepalive:
                self._keepalive_fut = asyncio.ensure_future(self.keep_alive(keepalive), loop=self.loop)
            if not appliances:
                await self.subscribe_all()
            else:
                await self.subscribe_appliances(appliances)
            await self.get_appliance_list()
            await self.async_event(EVENT_CONNECTED, None)
            try:
                async for message in socket:
                    await self.process_message(message)
            except websockets.WebSocketException:
                _LOGGER.error("Unknown error reading socket")
        _LOGGER.info("Disconnected")
        if self._keepalive_fut is not None:
            self._keepalive_fut.cancel()
        await self.async_event(EVENT_DISCONNECTED, None)

    async def send_dict(self, msg_dict: Dict[str, Any]):
        """JSON encode a dictionary and send it."""
        payload = json.dumps(msg_dict)
        try:
            await self.websocket.send(payload)
        except websockets.ConnectionClosed:
            _LOGGER.info("Tried to send a message, but connection already closed.")

    async def keep_alive(self, keepalive: int = 30):
        """Send periodic pings to keep the connection alive."""
        while not self.websocket.closed:
            await asyncio.sleep(keepalive)
            _LOGGER.debug("Sending keepalive ping")
            await self.send_ping()

    async def subscribe_all(self):
        """Subscribe to all appliances."""
        msg_dict = {"kind": "websocket#subscribe", "action": "subscribe", "resources": ["/appliance/*/erd/*"]}
        await self.send_dict(msg_dict)

    async def subscribe_appliances(self, appliances: List[GeAppliance]):
        """Subscribe to a list of appliances."""
        msg_dict = {
            "kind": "websocket#subscribe",
            "action": "subscribe",
            "resources": [f"/appliance/{appliance.mac_addr}/erd/*" for appliance in appliances]
        }
        await self.send_dict(msg_dict)

    async def subscribe_appliance(self, appliance: GeAppliance):
        """Subscribe to a single appliance."""
        await self.subscribe_appliances([appliance])

    async def async_set_erd_value(self, appliance: GeAppliance, erd_code: ErdCodeType, erd_value: Any):
        if isinstance(erd_code, ErdCode):
            raw_erd_code = erd_code.value
        else:
            raw_erd_code = erd_code
        raw_erd_code = raw_erd_code.upper().replace("0X", "0x")

        mac_addr = appliance.mac_addr

        request_body = {
            "kind": "appliance#erdListEntry",
            "userId": self.user_id,
            "applianceId": appliance.mac_addr,
            "erd": raw_erd_code,
            "value": erd_value,
            "ackTimeout": 10,
            "delay": 0,
        }

        msg_dict = {
            "kind": "websocket#api",
            "action": "api",
            "host": API_HOST,
            "method": "POST",
            "path": f"/v1/appliance/{mac_addr}/erd/{raw_erd_code}",
            "id": f"{mac_addr}-{SET_ERD}-{raw_erd_code}",
            "body": request_body,
        }
        self._pending_erds[(mac_addr, raw_erd_code)] = erd_value
        await self.send_dict(msg_dict)

    async def async_request_update(self, appliance: GeAppliance):
        """Request an appliance send a full update."""
        _LOGGER.debug(f"Requesting update for client {appliance.mac_addr}")
        msg_dict = {
            "kind": "websocket#api",
            "action": "api",
            "host": API_HOST,
            "method": "GET",
            "path": f"/v1/appliance/{appliance.mac_addr}/erd",
            "id": f"{appliance.mac_addr}-{ALL_ERD}"
        }
        await self.send_dict(msg_dict)

    async def get_appliance_list(self):
        """Request the list of appliances on this account."""
        msg_dict = {
            "kind": "websocket#api",
            "action": "api",
            "host": API_HOST,
            "method": "GET",
            "path": "/v1/appliance",
            "id": LIST_APPLIANCES,
        }
        await self.send_dict(msg_dict)

    async def send_ping(self):
        """Send a ping."""
        msg_dict = {
            "kind": "websocket#ping",
            "id": "keepalive-ping",
            "action": "ping",
        }
        await self.send_dict(msg_dict)

    async def add_appliance(self, mac_addr: str, set_online: bool = True):
        """Add an appliance to the registry and request an update."""
        mac_addr = mac_addr.upper()
        if mac_addr in self.appliances:
            # TODO: Make this a package-specific exception
            raise RuntimeError('Trying to add duplicate appliance')
        new_appliance = GeAppliance(mac_addr, self)
        if set_online:
            new_appliance.set_available()

        _LOGGER.debug(f'Adding appliance {mac_addr}')
        self.appliances[mac_addr] = new_appliance
        await self.async_event(EVENT_ADD_APPLIANCE, new_appliance)
        await self.async_request_update(new_appliance)



from gekitchensdk.erd.converters.abstract import ErdReadWriteConverter, ErdReadOnlyConverter
from gekitchensdk.erd.erd_codes import ErdCodeType

def erd_decode_int(value: str) -> int:
    """Decode an integer value sent as a hex encoded string."""
    return int(value, 16)
def erd_encode_int(value: any, length: int = 2) -> str:
    """Encode an integer value as a hex string."""
    value = int(value)
    return value.to_bytes(length, 'big').hex()

class ErdIntConverter(ErdReadWriteConverter[int]):
    def __init__(self, erd_code: ErdCodeType = "Unknown", length: int = 2):
        super().__init__(erd_code)
        self.length = length
    def erd_decode(self, value: str) -> int:
        """Decode an integer value sent as a hex encoded string."""
        return erd_decode_int(value)
    def erd_encode(self, value) -> str:
        """Encode an integer value as a hex string."""
        return erd_encode_int(value)

class ErdReadOnlyIntConverter(ErdReadOnlyConverter[int]):
    def erd_decode(self, value: str) -> int:
        """Decode an integer value sent as a hex encoded string."""
        return erd_decode_int(value)



"""GE Kitchen Appliances SDK"""

__version__ = "0.3.2"


from .async_login_flow import (
    async_do_full_xmpp_flow,
    async_get_ge_token,
    async_get_mobile_device_token,
    async_get_oauth2_token,
    async_get_xmpp_credentials,
)
from .clients import *
from .const import *
from .erd import *
from .exception import *
from .ge_appliance import GeAppliance
from .login_flow import (
    do_full_xmpp_flow,
    get_ge_token,
    get_mobile_device_token,
    get_oauth2_token,
    get_xmpp_credentials,
)



from .erd_filter_status_converter import ErdFilterStatusConverter
from .fridge_door_status_converter import FridgeDoorStatusConverter
from .fridge_ice_bucket_status_converter import FridgeIceBucketStatusConverter
from .fridge_set_points_converter import FridgeSetPointsConverter
from .fridge_set_point_limits_converter import FridgeSetPointLimitsConverter
from .hot_water_status_converter import HotWaterStatusConverter
from .ice_maker_control_status_converter import IceMakerControlStatusConverter
from .fridge_model_info_converter import FridgeModelInfoConverter


import enum
from typing import Optional

from .const import *

@enum.unique
class ErdOvenState(enum.Enum):
    """
    Oven state constants for display purposes.
    These are found in ErdCurrentState.smali.  That they are not aligned with the values for
    ErdCookMode and AvailableCookMode is infuriating.
    """
    BAKE = 5
    BAKE_PREHEAT = 1
    BAKE_TWO_TEMP = 6
    BROIL_HIGH = 14
    BROIL_LOW = 13
    CLEAN_COOL_DOWN = 23
    CLEAN_STAGE1 = 21
    CLEAN_STAGE2 = 22
    CONV_BAKE = 7
    CONV_BAKE_PREHEAT = 2
    CONV_BAKE_TWO_TEMP = 8
    CONV_BROIL_CRISP = 17
    CONV_BROIL_HIGH = 15
    CONV_BROIL_LOW = 16
    CONV_MULTI_BAKE_PREHEAT = 3
    CONV_MULTI_TWO_BAKE = 10
    CONV_MUTLI_BAKE = 9
    CONV_ROAST = 11
    CONV_ROAST2 = 12
    CONV_ROAST_BAKE_PREHEAT = 4
    CUSTOM_CLEAN_STAGE2 = 24
    DELAY = 27
    NO_MODE = 0
    PROOF = 19
    SABBATH = 20
    STEAM_CLEAN_STAGE2 = 25
    STEAM_COOL_DOWN = 26
    WARM = 18
    OVEN_STATE_BAKE = "oven_bake"
    OVEN_STATE_BAKED_GOODS = "oven_state_baked_goods"
    OVEN_STATE_BROIL = "oven_state_broil"
    OVEN_STATE_CONV_BAKE = "oven_state_conv_bake"
    OVEN_STATE_CONV_BAKE_MULTI = "oven_state_conv_bake_multi"
    OVEN_STATE_CONV_BROIL = "oven_state_conv_broil"
    OVEN_STATE_CONV_ROAST = "oven_state_conv_roast"
    OVEN_STATE_DELAY_START = "oven_state_delay_start"
    OVEN_STATE_DUAL_BROIL_HIGH = "oven_state_dual_broil_high"
    OVEN_STATE_DUAL_BROIL_LOW = "oven_state_dual_broil_low"
    OVEN_STATE_FROZEN_PIZZA = "oven_state_frozen_pizza"
    OVEN_STATE_FROZEN_PIZZA_MULTI = "oven_state_frozen_pizza_multi"
    OVEN_STATE_FROZEN_SNACKS = "oven_state_frozen_snacks"
    OVEN_STATE_FROZEN_SNACKS_MULTI = "oven_state_frozen_snacks_multi"
    OVEN_STATE_PROOF = "oven_state_proof"
    OVEN_STATE_SELF_CLEAN = "oven_state_self_clean"
    OVEN_STATE_SPECIAL_X = "oven_state_speical_x"  # [sic.]
    OVEN_STATE_STEAM_START = "oven_state_steam_start"
    OVEN_STATE_WARM = "oven_state_warm"
    STATUS_DASH = "status_dash"

    def stringify(self, **kwargs) -> Optional[str]:
        from .oven_display_state_mapping import OVEN_DISPLAY_STATE_MAP
        return OVEN_DISPLAY_STATE_MAP.get(self, STATE_OVEN_UNKNOWN)
    def __str__(self) -> str:
        return self.stringify() or STATE_OVEN_UNKNOWN        


from ..abstract import ErdReadWriteConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import OvenCookSetting, ErdOvenCookMode, OVEN_COOK_MODE_MAP

class OvenCookModeConverter(ErdReadWriteConverter[OvenCookSetting]):
    def erd_decode(self, value: str) -> OvenCookSetting:
        """
        Get the cook mode and temperature.
        TODO: Figure out what the other 10 bytes are for.
            I'm guessing they have to do with two-temp cooking, probes, delayed starts, timers, etc.
        """
        byte_string = erd_decode_bytes(value)
        cook_mode_code = byte_string[0]
        temperature = int.from_bytes(byte_string[1:3], 'big')
        cook_mode = ErdOvenCookMode(cook_mode_code)
        return OvenCookSetting(cook_mode=OVEN_COOK_MODE_MAP[cook_mode], temperature=temperature, raw_bytes=byte_string)
    def erd_encode(self, value: OvenCookSetting) -> str:
        """Re-encode a cook mode and temperature
        TODO: Other ten bytes"""
        cook_mode = value.cook_mode
        cook_mode_code = OVEN_COOK_MODE_MAP.inverse[cook_mode].value
        cook_mode_hex = cook_mode_code.to_bytes(1, 'big').hex()
        temperature_hex = int(value.temperature).to_bytes(2, 'big').hex()
        return cook_mode_hex + temperature_hex + ('00' * 10)



import logging

from gekitchensdk.erd.converters.abstract import ErdReadOnlyConverter
from gekitchensdk.erd.converters.primitives import *
from gekitchensdk.erd.values.dishwasher import ErdRinseAgent, ErdRinseAgentRaw, RINSE_AGENT_RAW_MAP

_LOGGER = logging.getLogger(__name__)

class ErdRinseAgentConverter(ErdReadOnlyConverter[ErdRinseAgent]):
    def erd_decode(self, value: str) -> ErdRinseAgent:
        """ Decodes the dishwasher rinse agent status """
        try:
            raw = ErdRinseAgentRaw(value)
            _LOGGER.debug(f'raw rinse agent value: {raw}')
            return ErdRinseAgent(RINSE_AGENT_RAW_MAP[raw])
        except (ValueError, KeyError):
            return ErdRinseAgent.NA



from ..primitives import ErdReadOnlyStringConverter

class CycleNameConverter(ErdReadOnlyStringConverter):
    def erd_decode(self, value: str) -> str:
        name = super().erd_decode(value)
        if name.startswith("AutoSense"):
            name = "AutoSense"

        return name  



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.fridge import ErdFilterStatus

class ErdFilterStatusConverter(ErdReadOnlyConverter[ErdFilterStatus]):
    def erd_decode(self, value: str) -> ErdFilterStatus:
        """Decode water filter status.

        This appears to be 9 bytes, of which only the first two are obviously used. I suspect that the others
        relate to how much time remains on the filter.  Leaving as a TODO.
        """
        status_byte = value[:2]
        if status_byte == "00":
            status_byte = value[2:4]
        try:
            return ErdFilterStatus(status_byte)
        except ValueError:
            return ErdFilterStatus.NA



import enum
from logging import setLoggerClass
from typing import Optional

@enum.unique
class ErdDoorStatus(enum.Enum):
    """Fridge door status"""
    CLOSED = "00"
    OPEN = "01"
    NA = "FF"

    def boolify(self) -> Optional[bool]:
        if self.value == ErdDoorStatus.NA:
            return None
        return self.value == ErdDoorStatus.OPEN

    def stringify(self, **kwargs) -> Optional[str]:
        if self.value == ErdDoorStatus.NA:
            return None
        return self.name.title()



"""Constants"""

# OAuth2 credentials
OAUTH2_APP_ID = "com.ge.kitchen.wca.prd.android"
OAUTH2_CLIENT_ID = "564c31616c4f7474434b307435412b4d2f6e7672"
OAUTH2_CLIENT_SECRET = "6476512b5246446d452f697154444941387052645938466e5671746e5847593d"
OAUTH2_REDIRECT_URI = "brillion.4e617a766474657344444e562b5935566e51324a://oauth/redirect"

LOGIN_URL = "https://accounts.brillion.geappliances.com"
API_URL = "https://api.brillion.geappliances.com"
SECURE_URL = "https://secure.brillion.geappliances.com"

EVENT_ADD_APPLIANCE = "add_appliance"
EVENT_APPLIANCE_INITIAL_UPDATE = "appliance_got_type"
EVENT_APPLIANCE_STATE_CHANGE = "appliance_state_change"
EVENT_APPLIANCE_UPDATE_RECEIVED = "appliance_updated_received"
EVENT_CONNECTED = "connected"
EVENT_DISCONNECTED = "disconnected"
EVENT_GOT_APPLIANCE_LIST = "got_appliance_list"



from .ge_unsupported_operation_error import GeUnsupportedOperationError

class GeSetErdNotAllowedError(GeUnsupportedOperationError):
    """ Exception raised when ERD is not allowed to be set (readonly) """
    def __init__(self, erd_code: str, message="ERD cannot be set, it is readonly"):
        self.erd_code = erd_code
        self.message = message
    
    def __str__(self):
        return f'{self.erd_code} -> {self.message}'



"""Data model for GE kitchen appliances"""

import enum
import logging
from weakref import WeakValueDictionary
from typing import Any, Dict, Optional, Set, TYPE_CHECKING, Union
from slixmpp import JID

from .erd import ErdCode, ErdCodeType, ErdCodeClass, ErdApplianceType, ErdEncoder
from .exception import *

if TYPE_CHECKING:
    from .clients import GeBaseClient

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)


class GeAppliance:
    """Base class shared by all appliances"""

    # Registry of initialized appliances
    _appliance_cache = WeakValueDictionary()

    def __new__(cls, mac_addr: Union[str, JID], client: "GeBaseClient", *args, **kwargs):
        if isinstance(mac_addr, JID):
            mac_addr = str(mac_addr.user).split('_')[0]
        try:
            obj = cls._appliance_cache[mac_addr]  # type: "GeAppliance"
        except KeyError:
            obj = super(GeAppliance, cls).__new__(cls)
            obj.__init__(mac_addr, client)
            cls._appliance_cache[obj.mac_addr] = obj
            return obj
        else:
            if client.client_priority >= obj.client.client_priority:
                obj.client = client
        return obj

    def __init__(self, mac_addr: Union[str, JID], client: "GeBaseClient"):
        if isinstance(mac_addr, JID):
            mac_addr = str(mac_addr.user).split('_')[0]
        self._available = False
        self._mac_addr = mac_addr.upper()
        self._message_id = 0
        self._property_cache = {}  # type: Dict[ErdCodeType, Any]
        self.client = client
        self.initialized = False
        self._encoder = ErdEncoder()

    @property
    def mac_addr(self) -> str:
        return self._mac_addr.upper()

    @property
    def known_properties(self) -> Set[ErdCodeType]:
        return set(self._property_cache)

    async def async_request_update(self):
        """Request the appliance send a full state update"""
        await self.client.async_request_update(self)

    def set_available(self):
        _LOGGER.debug(f'{self.mac_addr} marked available')
        self._available = True

    def set_unavailable(self):
        _LOGGER.debug(f'{self.mac_addr} marked unavailable')
        self._available = False

    @property
    def available(self) -> bool:
        return self._available and self.initialized

    @property
    def appliance_type(self) -> Optional[ErdApplianceType]:
        return self._property_cache.get(ErdCode.APPLIANCE_TYPE)

    def translate_erd_code(self, erd_code: ErdCodeType) -> ErdCodeType:
        """
        Translates a code to it's native value or a string if it is not known.
        """
        return self._encoder.translate_code(erd_code)

    def decode_erd_value(self, erd_code: ErdCodeType, erd_value: str) -> Any:
        """
        Decode and ERD Code raw value into something useful.  If the erd_code is a string that
        cannot be resolved to a known ERD Code, the value will be treated as raw byte string.
        Unregistered ERD Codes will be translated as ints.

        :param erd_code: ErdCode or str, the ERD Code the value of which we want to decode
        :param erd_value: The raw ERD code value, usually a hex string without leading "0x"
        :return: The decoded value.
        """
        return self._encoder.decode_value(erd_code, erd_value)

    def encode_erd_value(self, erd_code: ErdCodeType, value: Any) -> str:
        """
        Encode an ERD Code value as a hex string.
        Only ERD Codes registered with self.erd_encoders will processed.  Otherwise an error will be returned.

        :param erd_code: ErdCode or str, the ERD Code the value of which we want to decode
        :param value: The value to re-encode
        :return: The encoded value as a hex string
        """
        return self._encoder.encode_value(erd_code, value)

    def get_erd_value(self, erd_code: ErdCodeType) -> Any:
        """
        Get the value of a property represented by an ERD Code
        :param erd_code: ErdCode or str, The ERD code for the property to get
        :return: The current cached value of that ERD code
        """
        erd_code = self._encoder.translate_code(erd_code)
        return self._property_cache[erd_code]

    def get_erd_code_class(self, erd_code: ErdCodeType) -> ErdCodeClass:
        """
        Get the classification for a given ErdCode
        """
        return self._encoder.get_code_class(erd_code)

    async def async_set_erd_value(self, erd_code: ErdCodeType, value: Any):
        """
        Send a new erd value to the appliance.
        :param erd_code: The ERD code to update
        :param value: The new value to set
        """
        erd_value = self.encode_erd_value(erd_code, value)
        await self.client.async_set_erd_value(self, erd_code, erd_value)

    def update_erd_value(
            self, erd_code: ErdCodeType, erd_value: str) -> bool:
        """
        Setter for ERD code values.

        :param erd_code: ERD code to update
        :param erd_value: The new value to set, as returned by the appliance (usually a hex string)
        :return: Boolean, True if the state changed, False if no value changed
        """
        erd_code = self._encoder.translate_code(erd_code)
        value = self.decode_erd_value(erd_code, erd_value)

        old_value = self._property_cache.get(erd_code)

        try:
            state_changed = ((old_value is None) != (value is None)) or (old_value != value)
        except ValueError:
            _LOGGER.info('Unable to compare new and prior states.')
            state_changed = False

        if state_changed:
            _LOGGER.debug(f'Setting {erd_code} to {value}')
        self._property_cache[erd_code] = value

        return state_changed

    def update_erd_values(self, erd_values: Dict[ErdCodeType, str]) -> Dict[ErdCodeType, Any]:
        """
        Set one or more ERD codes value at once

        :param erd_values: Dictionary of erd codes and their new values as raw hex strings
        :return: dictionary of new states
        """
        state_changes = {
            self._encoder.translate_code(k): self.decode_erd_value(k, v)
            for k, v in erd_values.items()
            if self.update_erd_value(k, v)
        }

        return state_changes

    def stringify_erd_value(self, value: Any, **kwargs) -> Optional[str]:
        """
        Stringifies a code value if possible.  If it can't be stringified, returns none.
        By default, enums are stringified using their title-cased name (after replacing 
        underscores)
        """

        try:
            if not value:
                return None

            stringify_op = getattr(value, "stringify", None)
            if callable(stringify_op):
                return value.stringify(**kwargs)
            elif isinstance(value, enum.Enum):
                return value.name.replace("_"," ").title()
            else:
                return str(value)
        except (ValueError, KeyError):
            return None

    def boolify_erd_value(self, value: Any) -> Optional[bool]:
        """
        Boolifies a code value if possible.  If it can't be boolified, returns none
        """

        try:
            if not value:
                return None
            if isinstance(value, bool):
                return value

            boolify_op = getattr(value, "boolify", None)
            if callable(boolify_op):
                return value.boolify()
            else:
                return None
        except:
            return None

    def __str__(self):
        appliance_type = self.appliance_type
        if appliance_type is None:
            appliance_type = 'Unknown Type'
        return f'{self.__class__.__name__}({self.mac_addr}) ({appliance_type})'

    def __format__(self, format_spec):
        return str(self).__format__(format_spec)



from ..abstract import ErdReadWriteConverter
from ..primitives import *

from gekitchensdk.erd.values import ErdClockFormat

class ErdClockFormatConverter(ErdReadWriteConverter[ErdClockFormat]):
    def erd_decode(self, value: str) -> ErdClockFormat:
        return ErdClockFormat(value)
    def erd_encode(self, value: ErdClockFormat) -> str:
        return value.value    



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Optional

from gekitchensdk.erd.erd_codes import ErdCodeType
from gekitchensdk.exception import GeSetErdNotAllowedError

T = TypeVar('T')

class ErdValueConverter(Generic[T], ABC):
    def __init__(self, erd_code: ErdCodeType = "Unknown", can_decode: bool = True, can_encode: bool = True):
        self.erd_code = erd_code
        self._can_decode = can_decode
        self._can_encode = can_encode

    @abstractmethod
    def erd_encode(self, value: T) -> str:
        pass
    @abstractmethod
    def erd_decode(self, value: str) -> T:
        pass
    @property
    def can_decode(self) -> bool:
        return self._can_decode
    @property
    def can_encode(self) -> bool:
        return self._can_encode

class ErdReadWriteConverter(Generic[T], ABC):
    def __init__(self, erd_code: ErdCodeType = "Unknown"):
        ErdValueConverter.__init__(self, erd_code, True, True)

class ErdReadOnlyConverter(ErdValueConverter[T]):
    def __init__(self, erd_code: ErdCodeType = "Unknown"):
        ErdValueConverter.__init__(self, erd_code, True, False)
    def erd_encode(self, value: T) -> str:
        raise GeSetErdNotAllowedError(self.erd_code)



import logging

from datetime import timedelta
from typing import Optional

from ..abstract import ErdReadWriteConverter, ErdReadOnlyConverter

_LOGGER = logging.getLogger(__name__)

def erd_decode_timespan(value: any) -> Optional[timedelta]:
    """ Decodes a raw integer as a time span, 65535 is treated as None. """
    minutes = int(value, 16)
    if minutes == 65535:
        _LOGGER.debug('Got timespan value of 65535. Treating as None.')
        return None
    return timedelta(minutes=minutes)
def erd_encode_timespan(value: Optional[timedelta]) -> str:
    """ Encodes a time span as an erd integer, None is encoded as 65535. """
    if value is None:
        minutes = 65535
    else:
        minutes = value.seconds // 60
    return minutes.to_bytes(4, 'big').hex()

class ErdTimeSpanConverter(ErdReadWriteConverter[Optional[timedelta]]):
    def erd_decode(self, value: str) -> Optional[timedelta]:
        """ Decodes a raw integer as a time span, 65535 is treated as None. """
        return erd_decode_timespan(value)
    def erd_encode(self, value: Optional[timedelta]) -> str:
        """ Encodes a time span as an erd integer, None is encoded as 65535. """
        return erd_encode_timespan(value)

class ErdReadOnlyTimeSpanConverter(ErdReadOnlyConverter[Optional[timedelta]]):
    def erd_decode(self, value: str) -> Optional[timedelta]:
        """ Decodes a raw integer as a time span, 65535 is treated as None. """
        return erd_decode_timespan(value)



import enum
from enum import auto

@enum.unique
class ErdCodeClass(enum.IntFlag):
    NONE = auto()
    GENERAL = auto()
    CLOCK = auto()
    TIMER = auto()
    COUNTER = auto()
    NON_ZERO_TEMPERATURE = auto()
    RAW_TEMPERATURE = auto()
    DOOR = auto()
    BATTERY = auto()
    LOCK_CONTROL = auto()
    SABBATH_CONTROL = auto()
    COOLING_CONTROL = auto()
    TEMPERATURE_CONTROL = auto()
    OVEN_SENSOR = auto()
    FRIDGE_SENSOR = auto()
    FREEZER_SENSOR = auto()
    DISPENSER_SENSOR = auto()
    DISHWASHER_SENSOR = auto() 
 


import enum
from .available_cook_mode import AvailableCookMode
from .erd_oven_cook_mode import ErdOvenCookMode

@enum.unique
class ErdAvailableCookMode(enum.Enum):
    """
    Available cooking modes.
    In the XMPP API, they are represented as an index into an array of bytes and a bitmask.
    Thus these take the form (byte: int, mask: int, cook_mode: ErdOvenCookMode).  See ErdAvailableCookMode.smali
    in the Android app.
    """
    OVEN_BAKE = AvailableCookMode(byte=9, mask=2, cook_mode=ErdOvenCookMode.BAKE_NOOPTION)
    OVEN_CONVECTION_BAKE = AvailableCookMode(byte=7, mask=4, cook_mode=ErdOvenCookMode.CONVBAKE_NOOPTION)
    OVEN_CONVECTION_MULTI_BAKE = AvailableCookMode(byte=6, mask=8, cook_mode=ErdOvenCookMode.CONVMULTIBAKE_NOOPTION)
    OVEN_CONVECTION_ROAST = AvailableCookMode(byte=5, mask=16, cook_mode=ErdOvenCookMode.CONVROAST_NOOPTION)
    OVEN_FROZEN_SNACKS = AvailableCookMode(byte=2, mask=1, cook_mode=ErdOvenCookMode.FROZEN_SNACKS)
    OVEN_FROZEN_SNACKS_MULTI = AvailableCookMode(byte=2, mask=2, cook_mode=ErdOvenCookMode.FROZEN_SNACKS_MULTI)
    OVEN_FROZEN_PIZZA = AvailableCookMode(byte=2, mask=4, cook_mode=ErdOvenCookMode.FROZEN_PIZZA)
    OVEN_FROZEN_PIZZA_MULTI = AvailableCookMode(byte=2, mask=8, cook_mode=ErdOvenCookMode.FROZEN_PIZZA_MULTI)
    OVEN_BAKED_GOODS = AvailableCookMode(byte=2, mask=16, cook_mode=ErdOvenCookMode.BAKED_GOODS)




from ..abstract import ErdReadOnlyConverter
from ..primitives import *

class ErdModelSerialConverter(ErdReadOnlyConverter[str]):
    def erd_decode(self, value) -> str:
        """
        Decode a serial/model number string value sent as a hex encoded string.

        TODO: I think the first byte is a checksum.  I need to confirm this so we can have an encoder as well.
        """
        raw_bytes = bytes.fromhex(value)
        raw_bytes = raw_bytes.rstrip(b'\x00')

        return raw_bytes[1:].decode('ascii')



from ..abstract import ErdReadWriteConverter
from ..primitives import *

from gekitchensdk.erd.values import ErdSoundLevel

class ErdSoundLevelConverter(ErdReadWriteConverter[ErdSoundLevel]):    
    def erd_decode(self, value: str) -> ErdSoundLevel:
        sound_level = erd_decode_int(value)
        return ErdSoundLevel(sound_level)
    def erd_encode(self, value: ErdSoundLevel) -> str:
        return erd_encode_int(value.value)



from .erd_cycle_state import ErdCycleStateRaw, ErdCycleState
from .erd_operating_mode import ErdOperatingMode
from .erd_rinse_agent import ErdRinseAgentRaw, ErdRinseAgent
from .operating_mode import OperatingMode
from .cycle_state_mapping import CYCLE_STATE_RAW_MAP
from .operating_mode_mapping import OPERATING_MODE_MAP
from .rinse_agent_mapping import RINSE_AGENT_RAW_MAP



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.fridge import FridgeIceBucketStatus, ErdFullNotFull

class FridgeIceBucketStatusConverter(ErdReadOnlyConverter[FridgeIceBucketStatus]):
    def erd_decode(self, value: str) -> FridgeIceBucketStatus:
        """Decode Ice bucket status"""
        if not value:
            n = 0
        else:
            n = ErdIntConverter.erd_decode(value)

        is_present_ff = bool(n & 1)
        is_present_fz = bool(n & 2)
        state_full_ff = ErdFullNotFull.FULL if n & 4 else ErdFullNotFull.NOT_FULL
        state_full_fz = ErdFullNotFull.FULL if n & 8 else ErdFullNotFull.NOT_FULL

        if not is_present_ff:
            state_full_ff = ErdFullNotFull.NA
        if not is_present_fz:
            state_full_fz = ErdFullNotFull.NA

        if not (is_present_ff or is_present_ff):
            # No ice buckets at all
            total_status = ErdFullNotFull.NA
        elif (state_full_ff == ErdFullNotFull.NOT_FULL) or (state_full_fz == ErdFullNotFull.NOT_FULL):
            # At least one bucket is not full
            total_status = ErdFullNotFull.NOT_FULL
        else:
            total_status = ErdFullNotFull.FULL

        ice_status = FridgeIceBucketStatus(
            state_full_fridge=state_full_ff,
            state_full_freezer=state_full_fz,
            is_present_fridge=is_present_ff,
            is_present_freezer=is_present_fz,
            total_status=total_status,
        )
        return ice_status



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import OvenRanges

class OvenRangesConverter(ErdReadOnlyConverter[OvenRanges]):  
    def erd_decode(self, value: str) -> OvenRanges:
        raw_bytes = bytes.fromhex(value)
        upper_temp = int.from_bytes(raw_bytes[:2], 'big')
        lower_temp = int.from_bytes(raw_bytes[-2:], 'big')
        return OvenRanges(
            lower=lower_temp, 
            upper=upper_temp
        )



from .erd_oven_state import ErdOvenState
from .const import *


OVEN_DISPLAY_STATE_MAP = {
    ErdOvenState.BAKE: STATE_OVEN_BAKE,
    ErdOvenState.BAKE_PREHEAT: STATE_OVEN_PREHEAT,
    ErdOvenState.BAKE_TWO_TEMP: STATE_OVEN_BAKE_TWO_TEMP,
    ErdOvenState.BROIL_HIGH: STATE_OVEN_BROIL_HIGH,
    ErdOvenState.BROIL_LOW: STATE_OVEN_BROIL_LOW,
    ErdOvenState.CLEAN_COOL_DOWN: STATE_OVEN_SELF_CLEAN,
    ErdOvenState.CLEAN_STAGE1: STATE_OVEN_SELF_CLEAN,
    ErdOvenState.CLEAN_STAGE2: STATE_OVEN_SELF_CLEAN,
    ErdOvenState.CONV_BAKE: STATE_OVEN_CONV_BAKE,
    ErdOvenState.CONV_BAKE_PREHEAT: STATE_OVEN_PREHEAT,
    ErdOvenState.CONV_BAKE_TWO_TEMP: STATE_OVEN_CONV_BAKE_TWO_TEMP,
    ErdOvenState.CONV_BROIL_CRISP: STATE_OVEN_CONV_BROIL_CRISP,
    ErdOvenState.CONV_BROIL_HIGH: STATE_OVEN_CONV_BROIL_HIGH,
    ErdOvenState.CONV_BROIL_LOW: STATE_OVEN_CONV_BROIL_LOW,
    ErdOvenState.CONV_MULTI_BAKE_PREHEAT: STATE_OVEN_PREHEAT,
    ErdOvenState.CONV_MULTI_TWO_BAKE: STATE_OVEN_MULTI_BAKE,
    ErdOvenState.CONV_MUTLI_BAKE: STATE_OVEN_MULTI_BAKE,
    ErdOvenState.CONV_ROAST: STATE_OVEN_CONV_ROAST,
    ErdOvenState.CONV_ROAST2: STATE_OVEN_CONV_ROAST_TWO_TEMP,
    ErdOvenState.CONV_ROAST_BAKE_PREHEAT: STATE_OVEN_PREHEAT,
    ErdOvenState.CUSTOM_CLEAN_STAGE2: STATE_OVEN_SELF_CLEAN,
    ErdOvenState.DELAY: STATE_OVEN_DELAY,
    ErdOvenState.NO_MODE: STATE_OVEN_OFF,
    ErdOvenState.PROOF: STATE_OVEN_PROOF,
    ErdOvenState.SABBATH: STATE_OVEN_SABBATH,
    ErdOvenState.STEAM_CLEAN_STAGE2: STATE_OVEN_STEAM_CLEAN,
    ErdOvenState.STEAM_COOL_DOWN: STATE_OVEN_STEAM_CLEAN,
    ErdOvenState.WARM: STATE_OVEN_WARM,
    ErdOvenState.OVEN_STATE_BAKE: STATE_OVEN_BAKE,
    ErdOvenState.OVEN_STATE_BAKED_GOODS: STATE_OVEN_BAKED_GOODS,
    ErdOvenState.OVEN_STATE_BROIL: STATE_OVEN_BROIL_HIGH,
    ErdOvenState.OVEN_STATE_CONV_BAKE: STATE_OVEN_CONV_BAKE,
    ErdOvenState.OVEN_STATE_CONV_BAKE_MULTI: STATE_OVEN_CONV_BAKE_MULTI,
    ErdOvenState.OVEN_STATE_CONV_BROIL: STATE_OVEN_CONV_BROIL_HIGH,
    ErdOvenState.OVEN_STATE_CONV_ROAST: STATE_OVEN_CONV_ROAST,
    ErdOvenState.OVEN_STATE_DUAL_BROIL_HIGH: STATE_OVEN_DUAL_BROIL_HIGH,
    ErdOvenState.OVEN_STATE_DUAL_BROIL_LOW: STATE_OVEN_DUAL_BROIL_LOW,
    ErdOvenState.OVEN_STATE_DELAY_START: STATE_OVEN_DELAY,
    ErdOvenState.OVEN_STATE_FROZEN_PIZZA: STATE_OVEN_FROZEN_PIZZA,
    ErdOvenState.OVEN_STATE_FROZEN_PIZZA_MULTI: STATE_OVEN_FROZEN_PIZZA,
    ErdOvenState.OVEN_STATE_FROZEN_SNACKS: STATE_OVEN_FROZEN_SNACKS,
    ErdOvenState.OVEN_STATE_FROZEN_SNACKS_MULTI: STATE_OVEN_FROZEN_SNACKS,
    ErdOvenState.OVEN_STATE_PROOF: STATE_OVEN_PROOF,
    ErdOvenState.OVEN_STATE_SELF_CLEAN: STATE_OVEN_SELF_CLEAN,
    ErdOvenState.OVEN_STATE_SPECIAL_X: STATE_OVEN_SPECIAL,
    ErdOvenState.OVEN_STATE_STEAM_START: STATE_OVEN_STEAM_CLEAN,
    ErdOvenState.OVEN_STATE_WARM: STATE_OVEN_WARM,
    ErdOvenState.STATUS_DASH: STATE_OVEN_OFF,
}



import logging

from gekitchensdk.erd.converters.abstract import ErdReadOnlyConverter
from gekitchensdk.erd.converters.primitives import *
from gekitchensdk.erd.values.dishwasher import ErdCycleState, ErdCycleStateRaw, CYCLE_STATE_RAW_MAP

_LOGGER = logging.getLogger(__name__)

class ErdCycleStateConverter(ErdReadOnlyConverter[ErdCycleState]):
    def erd_decode(self, value: str) -> ErdCycleState:
        """ Decodes the dishwasher cycle state """    
        try:
            raw = ErdCycleStateRaw(value)
            _LOGGER.debug(f'raw cycle state value: {raw}')
            return ErdCycleState(CYCLE_STATE_RAW_MAP[raw])
        except (ValueError, KeyError):
            return ErdCycleState.NA



from ..abstract import ErdReadWriteConverter
from ..primitives import *
from gekitchensdk.erd.values.fridge import FridgeSetPoints

class FridgeSetPointsConverter(ErdReadWriteConverter[FridgeSetPoints]):
    def erd_decode(self, value: str) -> FridgeSetPoints:
        return FridgeSetPoints(
            fridge=ErdSignedByteConverter.erd_decode(value[0:2]),
            freezer=ErdSignedByteConverter.erd_decode(value[2:4]),
        )
    @staticmethod    
    def erd_encode(value: FridgeSetPoints):
        return ErdSignedByteConverter.erd_encode(value.fridge) + ErdSignedByteConverter.erd_encode(value.freezer)



import bidict

from .erd_oven_cook_mode import ErdOvenCookMode
from .erd_oven_state import ErdOvenState
from .oven_cook_mode import OvenCookMode

# Translate OVEN_COOK_MODE values into something easier to work with
OVEN_COOK_MODE_MAP = bidict.bidict({
    ErdOvenCookMode.BAKED_GOODS: OvenCookMode(ErdOvenState.OVEN_STATE_BAKED_GOODS, False, False, False),
    ErdOvenCookMode.BAKETIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.BAKE, True, True, False),
    ErdOvenCookMode.BAKETIMED_TWOTEMP: OvenCookMode(ErdOvenState.BAKE_TWO_TEMP, False, True, False),
    ErdOvenCookMode.BAKETIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.BAKE_TWO_TEMP, True, True, False),
    ErdOvenCookMode.BAKETIMED_WARM: OvenCookMode(ErdOvenState.WARM, False, True, False),
    ErdOvenCookMode.BAKETIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.WARM, True, True, False),
    ErdOvenCookMode.BAKE_DELAYSTART: OvenCookMode(ErdOvenState.BAKE, True, False, False),
    ErdOvenCookMode.BAKE_NOOPTION: OvenCookMode(ErdOvenState.BAKE, False, False, False),
    ErdOvenCookMode.BAKE_PROBE: OvenCookMode(ErdOvenState.BAKE, False, False, True),
    ErdOvenCookMode.BAKE_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.BAKE, True, False, True),
    ErdOvenCookMode.BAKE_SABBATH: OvenCookMode(ErdOvenState.BAKE, False, False, False, False, True),
    ErdOvenCookMode.BROIL_HIGH: OvenCookMode(ErdOvenState.BROIL_HIGH, False, False, False),
    ErdOvenCookMode.BROIL_LOW: OvenCookMode(ErdOvenState.BROIL_LOW, False, False, False),
    ErdOvenCookMode.CONVBAKETIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, True, False),
    ErdOvenCookMode.CONVBAKETIMED_TWOTEMP: OvenCookMode(ErdOvenState.CONV_BAKE_TWO_TEMP, False, True, False),
    ErdOvenCookMode.CONVBAKETIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE_TWO_TEMP, True, True, False),
    ErdOvenCookMode.CONVBAKETIMED_WARM: OvenCookMode(ErdOvenState.CONV_BAKE, False, True, False, True),
    ErdOvenCookMode.CONVBAKETIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, True, False, True),
    ErdOvenCookMode.CONVBAKE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, False, False),
    ErdOvenCookMode.CONVBAKE_NOOPTION: OvenCookMode(ErdOvenState.CONV_BAKE, False, False, False),
    ErdOvenCookMode.CONVBAKE_PROBE: OvenCookMode(ErdOvenState.CONV_BAKE, False, False, True),
    ErdOvenCookMode.CONVBAKE_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, False, True),
    ErdOvenCookMode.CONVBROILCRISP_NOOPTION: OvenCookMode(ErdOvenState.CONV_BROIL_CRISP, False, False, False),
    ErdOvenCookMode.CONVBROILCRISP_PROBE: OvenCookMode(ErdOvenState.CONV_BROIL_CRISP, False, False, True),
    ErdOvenCookMode.CONVBROIL_HIGH_NOOPTION: OvenCookMode(ErdOvenState.CONV_BROIL_HIGH, False, False, False),
    ErdOvenCookMode.CONVBROIL_LOW_NOOPTION: OvenCookMode(ErdOvenState.CONV_BROIL_LOW, False, False, False),
    ErdOvenCookMode.CONVMULTIBAKETIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, True, False),
    ErdOvenCookMode.CONVMULTIBAKETIMED_TWOTEMP: OvenCookMode(ErdOvenState.CONV_MULTI_TWO_BAKE, False, True, False),
    ErdOvenCookMode.CONVMULTIBAKETIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MULTI_TWO_BAKE, True, True, False),
    ErdOvenCookMode.CONVMULTIBAKETIMED_WARM: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, False, True, False, True),
    ErdOvenCookMode.CONVMULTIBAKETIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, True, False, True),
    ErdOvenCookMode.CONVMULTIBAKE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, False, False),
    ErdOvenCookMode.CONVMULTIBAKE_NOOPTION: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, False, False, False),
    ErdOvenCookMode.CONVMULTIBAKE_PROBE: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, False, False, True),
    ErdOvenCookMode.CONVMULTIBAKE_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, False, True),
    ErdOvenCookMode.CONVROASTTIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, True, False),
    ErdOvenCookMode.CONVROASTTIMED_TWOTEMP: OvenCookMode(ErdOvenState.CONV_ROAST2, False, True, False),
    ErdOvenCookMode.CONVROASTTIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST2, True, True, False),
    ErdOvenCookMode.CONVROASTTIMED_WARM: OvenCookMode(ErdOvenState.CONV_ROAST, False, True, False, True),
    ErdOvenCookMode.CONVROASTTIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, True, False, True),
    ErdOvenCookMode.CONVROAST_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, False, False),
    ErdOvenCookMode.CONVROAST_NOOPTION: OvenCookMode(ErdOvenState.CONV_ROAST, False, False, False),
    ErdOvenCookMode.CONVROAST_PROBE: OvenCookMode(ErdOvenState.CONV_ROAST, False, False, True),
    ErdOvenCookMode.CONVROAST_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, False, True),
    ErdOvenCookMode.CUSTOMSELFCLEAN: OvenCookMode(ErdOvenState.CUSTOM_CLEAN_STAGE2, False, False, False),
    ErdOvenCookMode.CUSTOMSELFCLEAN_DELAYSTART: OvenCookMode(ErdOvenState.CUSTOM_CLEAN_STAGE2, True, False, False),
    ErdOvenCookMode.DUALBROIL_HIGH_NOOPTION: OvenCookMode(ErdOvenState.OVEN_STATE_DUAL_BROIL_HIGH, False, False, False),
    ErdOvenCookMode.DUALBROIL_LOW_NOOPTION: OvenCookMode(ErdOvenState.OVEN_STATE_DUAL_BROIL_LOW, False, False, False),
    ErdOvenCookMode.FROZEN_PIZZA: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_PIZZA, False, False, False),
    ErdOvenCookMode.FROZEN_PIZZA_MULTI: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_PIZZA_MULTI, False, False, False),
    ErdOvenCookMode.FROZEN_SNACKS: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_SNACKS, False, False, False),
    ErdOvenCookMode.FROZEN_SNACKS_MULTI: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_SNACKS_MULTI, False, False, False),
    ErdOvenCookMode.NOMODE: OvenCookMode(ErdOvenState.NO_MODE, False, False, False),
    ErdOvenCookMode.PROOF_DELAYSTART: OvenCookMode(ErdOvenState.PROOF, True, False, False),
    ErdOvenCookMode.PROOF_NOOPTION: OvenCookMode(ErdOvenState.PROOF, False, False, False),
    ErdOvenCookMode.STEAMCLEAN: OvenCookMode(ErdOvenState.STEAM_CLEAN_STAGE2, False, False, False),
    ErdOvenCookMode.STEAMCLEAN_DELAYSTART: OvenCookMode(ErdOvenState.STEAM_CLEAN_STAGE2, True, False, False),
    ErdOvenCookMode.WARM_DELAYSTART: OvenCookMode(ErdOvenState.WARM, True, False, False),
    ErdOvenCookMode.WARM_NOOPTION: OvenCookMode(ErdOvenState.WARM, False, False, False),
    ErdOvenCookMode.WARM_PROBE: OvenCookMode(ErdOvenState.WARM, False, False, True),
})


from ..abstract import ErdReadOnlyConverter, ErdReadWriteConverter

def erd_decode_signed_byte(value: any) -> int:
    """
    Convert a hex byte to a signed int.  Copied from GE's hextodec method.
    """
    val = int(value, 16)
    if val > 128:
        return val - 256
    return val
def erd_encode_signed_byte(value: int) -> str:
    """
    Convert a hex byte to a signed int.  Copied from GE's hextodec method.
    """
    value = int(value)
    if value < 0:
        value = value + 256
    return value.to_bytes(1, "big").hex()

class ErdSignedByteConverter(ErdReadWriteConverter[int]):
    def erd_decode(self, value: str) -> int:
        """
        Convert a hex byte to a signed int.  Copied from GE's hextodec method.
        """
        return erd_decode_signed_byte(value)
    def erd_encode(self, value: int) -> str:
        """
        Convert a hex byte to a signed int.  Copied from GE's hextodec method.
        """
        return erd_encode_signed_byte(value)

class ErdReadOnlySignedByteConverter(ErdReadOnlyConverter[int]):
    def erd_decode(self, value: str) -> int:
        """
        Convert a hex byte to a signed int.  Copied from GE's hextodec method.
        """
        return erd_decode_signed_byte(value)



"""Base client for GE ERD APIs"""

import abc
import asyncio
import logging
from aiohttp import ClientSession
from typing import Any, Callable, Dict, Optional, Tuple
from ..const import EVENT_APPLIANCE_INITIAL_UPDATE
from ..erd import ErdCode, ErdCodeType, ErdEncoder
from ..exception import GeNotAuthenticatedError
from ..ge_appliance import GeAppliance

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)


class GeBaseClient(metaclass=abc.ABCMeta):
    """Abstract base class for GE ERD APIs"""

    client_priority = 0  # Priority of this client class.  Higher is better.

    def __init__(self, event_loop: Optional[asyncio.AbstractEventLoop] = None):
        self._loop = event_loop
        self._appliances = {}  # type: Dict[str, GeAppliance]
        self._credentials = None  # type: Optional[Dict]

    @property
    def credentials(self) -> Optional[Dict]:
        return self._credentials

    @credentials.setter
    def credentials(self, credentials: Dict):
        self._credentials = credentials

    @property
    def appliances(self) -> Dict[str, GeAppliance]:
        return self._appliances

    @property
    def user_id(self) -> Optional[str]:
        try:
            return self.credentials['userId']
        except (TypeError, KeyError):
            raise GeNotAuthenticatedError

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    @abc.abstractmethod
    async def async_set_erd_value(self, appliance: GeAppliance, erd_code: ErdCodeType, erd_value: Any):
        """
        Send a new erd value to the appliance
        :param appliance: The appliance being updated
        :param erd_code: The ERD code to update
        :param erd_value: The new value to set
        """
        pass

    @abc.abstractmethod
    async def async_request_update(self, appliance: GeAppliance):
        """Request the appliance send a full state update"""
        pass

    @abc.abstractmethod
    async def async_get_credentials(self, session: ClientSession, username: str, password: str):
        """Get updated credentials"""
        pass

    @abc.abstractmethod
    async def async_event(self, event: str, data: Any = None):
        """Trigger an event."""
        pass

    @abc.abstractmethod
    def add_event_handler(self, event: str, callback: Callable, disposable: bool):
        pass

    async def maybe_trigger_appliance_init_event(self, data: Tuple[GeAppliance, Dict[ErdCodeType, Any]]):
        """
        Trigger the appliance_got_type event if appropriate

        :param data: GeAppliance updated and the updates
        """
        appliance, state_changes = data
        if ErdCode.APPLIANCE_TYPE in state_changes and not appliance.initialized:
            _LOGGER.debug(f'Got initial appliance type for {appliance:s}')
            appliance.initialized = True
            await self.async_event(EVENT_APPLIANCE_INITIAL_UPDATE, appliance)



"""ERD Codes for GE appliances"""
import enum
from typing import Union

@enum.unique
class ErdCode(enum.Enum):
    """
    ERD codes for GE kitchen appliance properties.
    These were mostly lifted from ERD.smali in the the GE SmartHQ app v1.0.3.13
    """
    APPLIANCE_TYPE = "0x0008"
    CLOCK_FORMAT = "0x0006"
    CLOCK_TIME = "0x0005"
    MODEL_NUMBER = "0x0001"
    SABBATH_MODE = "0x0009"
    SERIAL_NUMBER = "0x0002"
    SOUND_LEVEL = "0x000a"
    TEMPERATURE_UNIT = "0x0007"
    USER_INTERFACE_LOCKED = "0x0004"

    # Low-level-type things
    WIFI_MODULE_SW_VERSION = "0x0100"
    WIFI_MODULE_SW_VERSION_AVAILABLE = "0x0101"
    ACM_UPDATING = "0x0102"
    APPLIANCE_SW_VERSION = "0x0103"
    APPLIANCE_SW_VERSION_AVAILABLE = "0x0104"
    APPLIANCE_UPDATING = "0x0105"
    LCD_SW_VERSION = "0x0106"
    LCD_SW_VERSION_AVAILABLE = "0x0107"
    LCD_UPDATING = "0x0108"

    # Dishwasher Codes
    CYCLE_NAME = "0x301c"
    CYCLE_STATE = "0x300e"
    OPERATING_MODE = "0x3001"
    PODS_REMAINING_VALUE = "0x301f"
    RINSE_AGENT = "0x3003"
    SOUND = "0x3007"
    TIME_REMAINING = "0xd004"

    # Fridge codes
    AIR_FILTER_STATUS = "0x101c"
    DOOR_STATUS = "0x1016"
    FRIDGE_MODEL_INFO = "0x101d"
    HOT_WATER_IN_USE = "0x1018"
    HOT_WATER_SET_TEMP = "0x1011"
    HOT_WATER_STATUS = "0x1010"
    ICE_MAKER_BUCKET_STATUS = "0x1007"
    ICE_MAKER_CONTROL = "0x100a"
    SETPOINT_LIMITS = "0x100b"
    CURRENT_TEMPERATURE = "0x1004"
    TEMPERATURE_SETTING = "0x1005"
    TURBO_COOL_STATUS = "0x100f"
    TURBO_FREEZE_STATUS = "0x100e"
    WATER_FILTER_STATUS = "0x1009"

    # Oven codes
    ACTIVE_F_CODE_STATUS = "0x5005"
    CONVECTION_CONVERSION = "0x5003"
    ELAPSED_ON_TIME = "0x5004"
    END_TONE = "0x5001"
    HOUR_12_SHUTOFF_ENABLED = "0x5000"
    KEY_PRESSED = "0x5006"
    LIGHT_BAR = "0x5002"
    LOWER_OVEN_AVAILABLE_COOK_MODES = "0x520b"
    LOWER_OVEN_COOK_MODE = "0x5200"
    LOWER_OVEN_COOK_TIME_REMAINING = "0x5204"
    LOWER_OVEN_CURRENT_STATE = "0x5201"
    LOWER_OVEN_DELAY_TIME_REMAINING = "0x5202"
    LOWER_OVEN_DISPLAY_TEMPERATURE = "0x5209"
    LOWER_OVEN_ELAPSED_COOK_TIME = "0x5208"
    LOWER_OVEN_KITCHEN_TIMER = "0x5205"
    LOWER_OVEN_PROBE_DISPLAY_TEMP = "0x5203"
    LOWER_OVEN_PROBE_PRESENT = "0x5207"
    LOWER_OVEN_REMOTE_ENABLED = "0x520a"
    LOWER_OVEN_USER_TEMP_OFFSET = "0x5206"
    LOWER_OVEN_WARMING_DRAWER_STATE = "0x520c"
    LOWER_OVEN_RAW_TEMPERATURE = "0x520d"
    OVEN_CONFIGURATION = "0x5007"
    OVEN_MODE_MIN_MAX_TEMP = "0x5008"
    UPPER_OVEN_AVAILABLE_COOK_MODES = "0x510b"
    UPPER_OVEN_COOK_MODE = "0x5100"
    UPPER_OVEN_COOK_TIME_REMAINING = "0x5104"
    UPPER_OVEN_CURRENT_STATE = "0x5101"
    UPPER_OVEN_DELAY_TIME_REMAINING = "0x5102"
    UPPER_OVEN_DISPLAY_TEMPERATURE = "0x5109"
    UPPER_OVEN_ELAPSED_COOK_TIME = "0x5108"
    UPPER_OVEN_KITCHEN_TIMER = "0x5105"
    UPPER_OVEN_PROBE_DISPLAY_TEMP = "0x5103"
    UPPER_OVEN_PROBE_PRESENT = "0x5107"
    UPPER_OVEN_REMOTE_ENABLED = "0x510a"
    UPPER_OVEN_USER_TEMP_OFFSET = "0x5106"
    UPPER_OVEN_WARMING_DRAWER_STATE = "0x510c"
    UPPER_OVEN_RAW_TEMPERATURE = "0x510d"
    WARMING_DRAWER_STATE = "0x5009"

    COOKTOP_CONFIG = "0x551c"
    COOKTOP_STATUS = "0x5520"

    PRECISION_COOKING_PROBE_CONTROL_MODE = "0x5670"
    PRECISION_COOKING_PROBE_STATUS = "0x5671"
    PRECISION_COOKING_PROBE_TEMP_TARGET = "0x5672" #R/W, int 4 places
    PRECISION_COOKING_PROBE_TEMP_CURRENT = "0x5673"
    PRECISION_COOKING_PROBE_TIME_TARGET = "0x5674" #R/W, int 4 places
    PRECISION_COOKING_START_SOUS_VIDE_TIMER_ACTIVE_STATUS = "0x5675" #R/W, int 2 places
    PRECISION_COOKING_PROBE_TIME_CURRENT = "0x5676"
    PRECISION_COOKING_PROBE_TARGET_TIME_REACHED = "0x5677"
    PRECISION_COOKING_PROBE_BATTERY_STATUS = "0x5678"

    CLOSED_LOOP_COOKING_CONFIGURATION = "0x5770"

    # Microwave
    MICROWAVE_RECIPE_STATUS = "0x5300"
    # Lots more in 0x5Cxx

    # Hood

ErdCodeType = Union[ErdCode, str]


from .erd_door_status import ErdDoorStatus
from .erd_filter_status import ErdFilterStatus
from .erd_full_not_full import ErdFullNotFull
from .erd_hot_water_status import ErdHotWaterStatus
from .erd_pod_status import ErdPodStatus
from .fridge_door_status import FridgeDoorStatus
from .fridge_ice_bucket_status import FridgeIceBucketStatus
from .fridge_set_point_limits import FridgeSetPointLimits
from .fridge_set_points import FridgeSetPoints
from .hot_water_status import HotWaterStatus
from .ice_maker_control_status import IceMakerControlStatus
from .fridge_model_info import FridgeModelInfo


from .erd_cycle_state_converter import ErdCycleStateConverter
from .erd_rinse_agent_converter import ErdRinseAgentConverter
from .cycle_name_converter import CycleNameConverter
from .operating_mode_converter import OperatingModeConverter



import enum

# Note: From what I see through the code, the only check that is done
# is whether the config is non-zero.  We'll just coalesce all other
# values to present for now.
@enum.unique
class ErdCooktopConfig(enum.Enum):
    NONE = 0
    PRESENT = 1
    UNKNOWN = 255



import enum

# This doesn't actually appear to be used anywhere in the app, so not
# sure which is Yes or No...
@enum.unique
class ErdPrecisionCookingProbeTargetTimeReached(enum.Enum):
    NA  = -1
    NO  = 0
    YES = 1



from datetime import timedelta

from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.common import ErdPresent
from gekitchensdk.erd.values.fridge import HotWaterStatus, ErdHotWaterStatus, ErdFullNotFull, ErdPodStatus

class HotWaterStatusConverter(ErdReadOnlyConverter[HotWaterStatus]):
    def erd_decode(self, value: str) -> HotWaterStatus:
        if not value:
            return HotWaterStatus(
                status=ErdHotWaterStatus.NA,
                time_until_ready=None,
                current_temp=None,
                tank_full=ErdFullNotFull.NA,
                brew_module=ErdPresent.NA,
                pod_status=ErdPodStatus.NA,
            )
        try:
            status = ErdHotWaterStatus(value[:2])
        except ValueError:
            status = ErdHotWaterStatus.NA

        time_until_ready = timedelta(minutes=ErdIntConverter.erd_decode(value[2:6]))
        current_temp = ErdIntConverter.erd_decode(value[6:8])

        try:
            tank_full = ErdFullNotFull(value[8:10])
        except ValueError:
            tank_full = ErdFullNotFull.NA

        try:
            brew_module = ErdPresent(value[10:12])
        except ValueError:
            brew_module = ErdPresent.NA

        try:
            pod_status = ErdPodStatus(value[12:14])
        except ValueError:
            pod_status = ErdPodStatus.NA

        return HotWaterStatus(
            status=status,
            time_until_ready=time_until_ready,
            current_temp=current_temp,
            tank_full=tank_full,
            brew_module=brew_module,
            pod_status=pod_status,
        )



"""HTTP authentication flow to get XMPP loggin credentials"""

try:
    import re2 as re
except ImportError:
    import re
import logging
import requests
from lxml import etree
from gekitchensdk.const import (
    API_URL,
    LOGIN_URL,
    OAUTH2_APP_ID,
    OAUTH2_CLIENT_ID,
    OAUTH2_CLIENT_SECRET,
    OAUTH2_REDIRECT_URI,
)
from .exception import GeAuthFailedError

from typing import Dict
from urllib.parse import urlparse, parse_qs

_LOGGER = logging.getLogger(__name__)


def get_oauth2_token(session: requests.Session, username: str, password: str):
    """Hackily get an oauth2 token until I can be bothered to do this correctly"""
    params = {
        'client_id': OAUTH2_CLIENT_ID,
        'response_type': 'code',
        'access_type': 'offline',
        'redirect_uri': OAUTH2_REDIRECT_URI,
    }

    r1 = session.get(f'{LOGIN_URL}/oauth2/auth', params=params)

    email_regex = (
        r'^\s*(\w+(?:(?:-\w+)|(?:\.\w+)|(?:\+\w+))*\@'
        r'[A-Za-z0-9]+(?:(?:\.|-)[A-Za-z0-9]+)*\.[A-Za-z0-9][A-Za-z0-9]+)\s*$'
    )
    clean_username = re.sub(email_regex, r'\1', username)

    etr = etree.HTML(r1.text)
    post_data = {
        i.attrib['name']: i.attrib['value']
        for i in etr.xpath("//form[@id = 'frmsignin']//input")
        if 'value' in i.keys()
    }
    post_data['username'] = clean_username
    post_data['password'] = password

    r2 = session.post(f'{LOGIN_URL}/oauth2/g_authenticate', data=post_data, allow_redirects=False)
    code = parse_qs(urlparse(r2.headers['Location']).query)['code'][0]

    r3 = session.post(
        f'{LOGIN_URL}/oauth2/token',
        data={
            'code': code,
            'client_id': OAUTH2_CLIENT_ID,
            'client_secret': OAUTH2_CLIENT_SECRET,
            'redirect_uri': OAUTH2_REDIRECT_URI,
            'grant_type': 'authorization_code',
        }, auth=(OAUTH2_CLIENT_ID, OAUTH2_CLIENT_SECRET)
    )
    oauth_token = r3.json()
    try:
        session.headers.update({'Authorization': 'Bearer ' + oauth_token['access_token']})
    except KeyError:
        # TODO: make this better
        raise GeAuthFailedError(f'Failed to get a token: {oauth_token}')
    return oauth_token


def get_mobile_device_token(session: requests.Session) -> str:
    """Get a mobile device token"""
    mdt_data = {
        'kind': 'mdt#login',
        'app': OAUTH2_APP_ID,
        'os': 'google_android'
    }
    r = session.post(f'{API_URL}/v1/mdt', json=mdt_data)
    results = r.json()
    try:
        return r.json()['mdt']
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a mobile device token: {results}')


def get_wss_credentials(session: requests.Session) -> Dict:
    """Get a mobile device token"""
    r = session.get(f'{API_URL}/v1/websocket')
    return r.json()


def get_ge_token(session: requests.Session, mobile_device_token: str) -> str:
    """Get the GE token that we can use to get XMPP credentials"""
    params = {
        'client_id': OAUTH2_CLIENT_ID,
        'client_secret': OAUTH2_CLIENT_SECRET,
        'mdt': mobile_device_token
    }
    r = session.post(f'{LOGIN_URL}/oauth2/getoken', params=params)
    results = r.json()
    try:
        return results['access_token']
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a GE token: {results}')


def get_xmpp_credentials(ge_token: str) -> Dict:
    """Get XMPP credentials"""
    r = requests.get(
        f'{API_URL}/v1/mdt/credentials',
        headers={'Authorization': f'Bearer {ge_token}'}
    )
    return r.json()


def do_full_xmpp_flow(username: str, password: str) -> Dict:
    """Perform a complete login flow, returning XMPP credentials."""
    login_session = requests.session()

    _LOGGER.debug('Getting oauth2 token')
    get_oauth2_token(login_session, username, password)

    _LOGGER.debug('Getting mobile device token')
    mdt = get_mobile_device_token(login_session)

    _LOGGER.debug('Getting GE token')
    ge_token = get_ge_token(login_session, mdt)

    _LOGGER.debug('Getting XMPP credentials')
    xmpp_credentials = get_xmpp_credentials(ge_token)

    return xmpp_credentials


def do_full_wss_flow(username: str, password: str) -> Dict:
    """Perform a complete login flow, returning WSS credentials."""
    login_session = requests.session()

    _LOGGER.debug('Getting oauth2 token')
    get_oauth2_token(login_session, username, password)

    _LOGGER.debug('Getting WSS credentials')
    wss_credentials = get_wss_credentials(login_session)

    return wss_credentials



#not sure where this is used, but I found these mappings
#0x5778 = burner1
#0x5779 = burner2
#0x577a = burner3
#0x577b = burner4
#0x577c = burner5
#0x577d = burner6
#0x577e = burner1 + burner2

## These are guesses as to the meaning of the bits
## It's really hard to tell from what the app is showing in the interface, most aren't used
#burner.a = exists
#burner.b = on_off_only
#burner.c = something to do with sensor cooking...
#burner.d = something to do with being on... 
#burner.e = synchronized
#burner.f = on
#burner.g = power_pct

class Burner:
    def __init__(self, value: int, power: int):
        self.exists = self.shift(value, 0)
        #they skip the first bit shift (never stored, so not sure what it is)
        self.on_off_only = self.shift(value, 2) #b
        self.c = self.shift(value, 3) #c
        self.d = self.shift(value, 4) #d
        self.synchronized = self.shift(value, 5) #e
        self.on = self.shift(value, 6) #f

        self.power_pct = (~int("10000000", 2)) & power #g
        
    def shift(self, value: int, shift: int) -> bool:
        return ((value & (1 << shift)) >> shift) == 1



from typing import NamedTuple, Optional

class OvenConfiguration(NamedTuple):
    """Cleaner representation of ErdOvenConfiguration"""
    has_knob: bool
    has_warming_drawer: bool
    has_light_bar: bool
    has_lower_oven: bool
    has_lower_oven_kitchen_timer: bool
    raw_value: Optional[str] = None




""" GE Kitchen Exceptions """

from .ge_exception import GeException
from .ge_unsupported_operation_error import GeUnsupportedOperationError
from .ge_set_erd_not_allowed_error import GeSetErdNotAllowedError
from .ge_auth_failed_error import GeAuthFailedError
from .ge_client_disconnected_error import GeClientDisconnectedError
from .ge_not_authenticated_error import GeNotAuthenticatedError
from .ge_general_server_error import GeGeneralServerError


from ..abstract import ErdReadWriteConverter
from ..primitives import *

from gekitchensdk.erd.values import ErdEndTone

class ErdEndToneConverter(ErdReadWriteConverter[ErdEndTone]):
    def erd_decode(self, value: str) -> ErdEndTone:
        try:
            return ErdEndTone(value)
        except ValueError:
            return ErdEndTone.NA
    def erd_encode(self, value: ErdEndTone) -> str:
        if value == ErdEndTone.NA:
            raise ValueError("Invalid EndTone value")
        return value.value



from typing import Optional
from ..abstract import ErdReadOnlyConverter, ErdReadWriteConverter

def erd_decode_bool(value: any) -> Optional[bool]:
    """ Decodes a raw value to a bool, FF is considered None """
    if value == "FF":
        return None
    return bool(int(value))
def erd_encode_bool(value: Optional[bool]) -> str:
    """ Encodes a raw value to a bool, None is encoded as FF """
    if value is None:
        return "FF"
    return "01" if value else "00"    

class ErdBoolConverter(ErdReadWriteConverter[Optional[bool]]):
    def erd_decode(self, value: str) -> Optional[bool]:
        """ Decodes a raw value to a bool, FF is considered None """
        return erd_decode_bool(value)
    def erd_encode(self, value: Optional[bool]) -> str:
        """ Encodes a raw value to a bool, None is encoded as FF """
        return erd_encode_bool(value)
class ErdReadOnlyBoolConverter(ErdReadOnlyConverter[Optional[bool]]):
    def erd_decode(self, value: str) -> Optional[bool]:
        """ Decodes a raw value to a bool, FF is considered None """
        return erd_decode_bool(value)



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.fridge import FridgeSetPointLimits

class FridgeSetPointLimitsConverter(ErdReadOnlyConverter[FridgeSetPointLimits]):
    def erd_decode(self, value: str) -> FridgeSetPointLimits:
        return FridgeSetPointLimits(
            fridge_min=ErdSignedByteConverter.erd_decode(value[0:2]),
            fridge_max=ErdSignedByteConverter.erd_decode(value[2:4]),
            freezer_min=ErdSignedByteConverter.erd_decode(value[4:6]),
            freezer_max=ErdSignedByteConverter.erd_decode(value[6:8]),
        )



from typing import NamedTuple, Optional
from .erd_door_status import ErdDoorStatus

class FridgeDoorStatus(NamedTuple):
    fridge_right: ErdDoorStatus
    fridge_left: ErdDoorStatus
    freezer: ErdDoorStatus
    drawer: ErdDoorStatus
    status: str

    def any_open(self) -> Optional[bool]:
        if self.status.lower().endswith("open"):
            return True
        return False    

    def stringify(self, **kwargs) -> Optional[str]:
        return self.status
    
    def __str__(self) -> str:
        return self.status or ""


from typing import Any
from .erd_code_class import ErdCodeClass
from .erd_codes import ErdCode
from .converters import *

class ErdConfigurationEntry:
    def __init__(self, erd_code: ErdCode, converter: ErdValueConverter, code_class: ErdCodeClass) -> None:
        super().__init__()
        self.erd_code = erd_code
        self.converter = converter
        self.code_class = code_class
        self.converter.erd_code = self.erd_code

    @property
    def can_decode(self) -> bool:
        return self.converter.can_decode
    @property
    def can_encode(self) -> bool:
        return self.converter.can_encode
        
    def erd_decode(self, value: str) -> Any:
        return self.converter.erd_decode(value)
    def erd_encode(self, value: Any) -> str:
        return self.converter.erd_encode(value)
        
_configuration = [
    #Universal
    ErdConfigurationEntry(ErdCode.APPLIANCE_TYPE, ErdApplianceTypeConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.MODEL_NUMBER, ErdModelSerialConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.SERIAL_NUMBER, ErdModelSerialConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.SABBATH_MODE, ErdBoolConverter(), ErdCodeClass.SABBATH_CONTROL),
    ErdConfigurationEntry(ErdCode.USER_INTERFACE_LOCKED, ErdBoolConverter(), ErdCodeClass.LOCK_CONTROL),
    ErdConfigurationEntry(ErdCode.ACM_UPDATING, ErdReadOnlyBoolConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.APPLIANCE_UPDATING, ErdReadOnlyBoolConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.LCD_UPDATING, ErdReadOnlyBoolConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.CLOCK_FORMAT, ErdClockFormatConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.END_TONE, ErdEndToneConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.SOUND_LEVEL, ErdSoundLevelConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.TEMPERATURE_UNIT, ErdMeasurementUnitsConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.APPLIANCE_SW_VERSION, ErdSoftwareVersionConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.APPLIANCE_SW_VERSION_AVAILABLE, ErdSoftwareVersionConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.LCD_SW_VERSION, ErdSoftwareVersionConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.LCD_SW_VERSION_AVAILABLE, ErdSoftwareVersionConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.WIFI_MODULE_SW_VERSION, ErdSoftwareVersionConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.WIFI_MODULE_SW_VERSION_AVAILABLE, ErdSoftwareVersionConverter(), ErdCodeClass.GENERAL),

    #Fridge
    ErdConfigurationEntry(ErdCode.HOT_WATER_SET_TEMP, ErdIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.HOT_WATER_IN_USE, ErdReadOnlyBoolConverter(), ErdCodeClass.DISPENSER_SENSOR),
    ErdConfigurationEntry(ErdCode.TURBO_FREEZE_STATUS, ErdBoolConverter(), ErdCodeClass.COOLING_CONTROL),
    ErdConfigurationEntry(ErdCode.TURBO_COOL_STATUS, ErdBoolConverter(), ErdCodeClass.COOLING_CONTROL),
    ErdConfigurationEntry(ErdCode.DOOR_STATUS, FridgeDoorStatusConverter(), ErdCodeClass.DOOR),
    ErdConfigurationEntry(ErdCode.HOT_WATER_STATUS, HotWaterStatusConverter(), ErdCodeClass.DISPENSER_SENSOR),
    ErdConfigurationEntry(ErdCode.ICE_MAKER_BUCKET_STATUS, FridgeIceBucketStatusConverter(), ErdCodeClass.FREEZER_SENSOR),
    ErdConfigurationEntry(ErdCode.ICE_MAKER_CONTROL, IceMakerControlStatusConverter(), ErdCodeClass.FREEZER_SENSOR),
    ErdConfigurationEntry(ErdCode.AIR_FILTER_STATUS, ErdFilterStatusConverter(), ErdCodeClass.FRIDGE_SENSOR),
    ErdConfigurationEntry(ErdCode.WATER_FILTER_STATUS, ErdFilterStatusConverter(), ErdCodeClass.FRIDGE_SENSOR),
    ErdConfigurationEntry(ErdCode.SETPOINT_LIMITS, FridgeSetPointLimitsConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.CURRENT_TEMPERATURE, FridgeSetPointsConverter(), ErdCodeClass.RAW_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.TEMPERATURE_SETTING, FridgeSetPointsConverter(), ErdCodeClass.RAW_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.FRIDGE_MODEL_INFO, FridgeModelInfoConverter(), ErdCodeClass.GENERAL),
    
    #Oven
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_PROBE_PRESENT, ErdReadOnlyBoolConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_REMOTE_ENABLED, ErdReadOnlyBoolConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_DISPLAY_TEMPERATURE, ErdReadOnlyIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_PROBE_DISPLAY_TEMP, ErdReadOnlyIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_RAW_TEMPERATURE, ErdReadOnlyIntConverter(), ErdCodeClass.RAW_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_USER_TEMP_OFFSET, ErdIntConverter(), ErdCodeClass.RAW_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_COOK_TIME_REMAINING, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_DELAY_TIME_REMAINING, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_ELAPSED_COOK_TIME, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_KITCHEN_TIMER, ErdTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_CURRENT_STATE, ErdOvenStateConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_AVAILABLE_COOK_MODES, ErdAvailableCookModeConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.LOWER_OVEN_COOK_MODE, OvenCookModeConverter(), ErdCodeClass.OVEN_SENSOR),

    ErdConfigurationEntry(ErdCode.UPPER_OVEN_PROBE_PRESENT, ErdReadOnlyBoolConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_REMOTE_ENABLED, ErdReadOnlyBoolConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_DISPLAY_TEMPERATURE, ErdReadOnlyIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_PROBE_DISPLAY_TEMP, ErdReadOnlyIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_RAW_TEMPERATURE, ErdReadOnlyIntConverter(), ErdCodeClass.RAW_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_USER_TEMP_OFFSET, ErdIntConverter(), ErdCodeClass.RAW_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_COOK_TIME_REMAINING, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_DELAY_TIME_REMAINING, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_ELAPSED_COOK_TIME, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_KITCHEN_TIMER, ErdTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_CURRENT_STATE, ErdOvenStateConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_AVAILABLE_COOK_MODES, ErdAvailableCookModeConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.UPPER_OVEN_COOK_MODE, OvenCookModeConverter(), ErdCodeClass.OVEN_SENSOR),

    ErdConfigurationEntry(ErdCode.CONVECTION_CONVERSION, ErdBoolConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.HOUR_12_SHUTOFF_ENABLED, ErdBoolConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.OVEN_CONFIGURATION, OvenConfigurationConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.OVEN_MODE_MIN_MAX_TEMP, OvenRangesConverter(), ErdCodeClass.RAW_TEMPERATURE),

    ErdConfigurationEntry(ErdCode.COOKTOP_CONFIG, ErdCooktopConfigConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.COOKTOP_STATUS, CooktopStatusConverter(), ErdCodeClass.GENERAL),

    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_CONTROL_MODE, ErdPrecisionCookingAppProbeControlModeConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_STATUS, ErdReadOnlyIntConverter(), ErdCodeClass.OVEN_SENSOR),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_TEMP_TARGET, ErdIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_TEMP_CURRENT, ErdReadOnlyIntConverter(), ErdCodeClass.NON_ZERO_TEMPERATURE),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_TIME_TARGET, ErdTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_START_SOUS_VIDE_TIMER_ACTIVE_STATUS, ErdPrecisionCookingStartSousVideTimerActiveStatusConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_TIME_CURRENT, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_TARGET_TIME_REACHED, ErdPrecisionCookingProbeTargetTimeReachedConverter(), ErdCodeClass.GENERAL),
    ErdConfigurationEntry(ErdCode.PRECISION_COOKING_PROBE_BATTERY_STATUS, ErdPrecisionCookingProbeBatteryStatusConverter(), ErdCodeClass.BATTERY),

    # Dishwasher
    ErdConfigurationEntry(ErdCode.CYCLE_NAME, CycleNameConverter(), ErdCodeClass.DISHWASHER_SENSOR),
    ErdConfigurationEntry(ErdCode.PODS_REMAINING_VALUE, ErdIntConverter(), ErdCodeClass.COUNTER),
    ErdConfigurationEntry(ErdCode.TIME_REMAINING, ErdReadOnlyTimeSpanConverter(), ErdCodeClass.TIMER),
    ErdConfigurationEntry(ErdCode.CYCLE_STATE, ErdCycleStateConverter(), ErdCodeClass.DISHWASHER_SENSOR),
    ErdConfigurationEntry(ErdCode.OPERATING_MODE, OperatingModeConverter(), ErdCodeClass.DISHWASHER_SENSOR),
    ErdConfigurationEntry(ErdCode.RINSE_AGENT, ErdRinseAgentConverter(), ErdCodeClass.DISHWASHER_SENSOR)
]



"""GE Client implementations"""

import logging
from .base_client import GeBaseClient
from .websocket_client import GeWebsocketClient

_LOGGER = logging.getLogger(__name__)

try:
    from .xmpp_client import GeXmppClient
except ImportError:
    _LOGGER.info("XMPP client not avaible.  You may need to install slximpp.")



from ..abstract import ErdReadOnlyConverter
from ..primitives import *
from gekitchensdk.erd.values.oven import ErdCooktopConfig


class ErdCooktopConfigConverter(ErdReadOnlyConverter[ErdCooktopConfig]):
    def erd_decode(self, value: str):       
        if value:
            try:
                present = erd_decode_int(value[:2])
                return ErdCooktopConfig.NONE if present == 0 else ErdCooktopConfig.PRESENT 
            except ValueError:
                return ErdCooktopConfig.NONE
        else:
            return ErdCooktopConfig.NONE



import enum
from typing import Optional

@enum.unique
class ErdFullNotFull(enum.Enum):
    FULL = "01"
    NOT_FULL = "00"
    NA = "NA"

    def boolify(self) -> Optional[bool]:
        if self.value == ErdFullNotFull.NA:
            return None
        return self.value == ErdFullNotFull.FULL


from typing import NamedTuple, Optional
from datetime import timedelta
from ..common import ErdPresent
from .erd_hot_water_status import ErdHotWaterStatus
from .erd_full_not_full import ErdFullNotFull
from .erd_pod_status import ErdPodStatus

class HotWaterStatus(NamedTuple):
    status: ErdHotWaterStatus
    time_until_ready: Optional[timedelta]
    current_temp: Optional[int]
    tank_full: ErdFullNotFull
    brew_module: ErdPresent
    pod_status: ErdPodStatus

    @property
    def faulted(self) -> Optional[bool]:
        if self.status in [ErdHotWaterStatus.FAULT_NEED_CLEARED, ErdHotWaterStatus.FAULT_LOCKED_OUT]:
            return True
        return False



# WIP

import enum

@enum.unique
class ErdClosedLoopCookingConfiguration(enum.Enum):
    MODE_NOT_SUPPORTED_CLOSED_LOOP_COOKTOP = -1   
    OLD_PRECISION_COOKING = 1
    NEW_CLOSED_LOOP_COOKTOP = 2
    MODE_NOT_SUPPORTED_ALL = 3



from ..abstract import ErdReadWriteConverter, ErdReadOnlyConverter

def erd_decode_string(value: str) -> str:
    """
    Decode an string value sent as a hex encoded string.
    """
    raw_bytes = bytes.fromhex(value)
    raw_bytes = raw_bytes.rstrip(b'\x00')

    return raw_bytes.decode('ascii')
def erd_encode_string(value: str) -> str:
    """
    Encode an string value to a hex encoded string.
    """
    raw_bytes = value.encode('ascii')
    return bytes.hex(raw_bytes)

class ErdStringConverter(ErdReadWriteConverter[str]):
    def erd_decode(self, value: str) -> str:
        """
        Decode an string value sent as a hex encoded string.
        """
        return erd_decode_string(value)
    def erd_encode(self, value: str) -> str:
        """
        Encode an string value to a hex encoded string.
        """
        return erd_encode_string(value)

class ErdReadOnlyStringConverter(ErdReadOnlyConverter[str]):
    def erd_decode(self, value: str) -> str:
        """
        Decode an string value sent as a hex encoded string.
        """
        return erd_decode_string(value)
        


from typing import NamedTuple, Optional
from .erd_full_not_full import ErdFullNotFull

class FridgeIceBucketStatus(NamedTuple):
    state_full_fridge: ErdFullNotFull
    state_full_freezer: ErdFullNotFull
    is_present_fridge: bool
    is_present_freezer: bool
    total_status: ErdFullNotFull
   
    def stringify(self, **kwargs) -> Optional[str]:     
        if self.total_status == ErdFullNotFull.FULL:
            return "Full"
        if self.total_status == ErdFullNotFull.NOT_FULL:
            return "Not Full"
        if self.total_status == ErdFullNotFull.NA:
            return "NA"
        return None
    def __str__(self) -> str:
        return self.stringify() or ""